/*  Copyright (C) 2013 Zoltán Nagy
 *  Created by Zoltán Nagy on 6/4/13.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __cxx_cont_binary_tree_h__
#define __cxx_cont_binary_tree_h__ 1


/*   std includes   */
#include <cstddef>
#include <iterator>
#include <memory>
#include <type_traits>


/*   project headers   */
#include <cxx/containers/cont-tree-iterator.h>
#include <cxx/containers/cont-tree-alloc.h>


namespace duct {
  namespace container {
    
    
    /*   Defination of the  binary_tree class   */
    template<typename _Tp, bool _Recycle = false>
    class binary_tree
    {
      /*   node allocator    */
      using node_allocator = __tree_allocator<_Tp, _Recycle>;
      using link = __tree_node_link;
      
    public:
      /*    Standard typedef   */
      using value_type = _Tp;
      using pointer = value_type*;
      using reference = value_type&;
      using const_pointer = const value_type*;
      using const_reference = const value_type&;
      using size_type = std::size_t;
      using allocator_type = pmr::memory_resource*;
      
      
      /*   default constructor   */
      binary_tree() : _M_alloc(nullptr) {
        _M_root = _M_alloc.create_root();
      }
      
      explicit binary_tree(allocator_type a)
      : _M_alloc(a) { _M_root = _M_alloc.create_root();}
      
      /*   move   */
      binary_tree(binary_tree&& x)
      : binary_tree{} {
        _M_alloc.swap(x._M_alloc);
        std::swap(_M_root, x._M_root);
      }
      
      binary_tree(std::allocator_arg_t, allocator_type u, binary_tree&& x)
      : binary_tree{u}
      {
        /*   if the accomulators are equivalent then a simple swap   */
        if(u->is_equal(x._M_alloc.resource())) {
          _M_alloc.swap(x._M_alloc);
          std::swap(_M_root, x._M_root);
        }
        /*   otherwise move by elements   */
        else _M_alloc.move_a_branch(x._M_root->left);
      }

      binary_tree(binary_tree&& x, allocator_type u) : binary_tree{u}
      {
        /*   if the accomulators are equivalent then a simple swap   */
        if(u->is_equal(x._M_alloc.resource())) {
          _M_alloc.swap(x._M_alloc);
          std::swap(_M_root, x._M_root);
        }
        /*   otherwise move by elements   */
        else _M_alloc.move_a_branch(x._M_root->left);
      }
      
      binary_tree& operator=(binary_tree&& x)
      {
        if(this != &x) {
          std::swap(_M_root, x._M_root);
          _M_alloc.swap(x._M_alloc);
        }
        
        return *this;
      }
      
      /*   copy   */
      binary_tree(std::allocator_arg_t, allocator_type u, const binary_tree& x)
      : _M_alloc{u} {
        _M_root = _M_alloc.create_root();
        _M_root->left = _M_alloc.copy_a_branch(x._M_root->left);
        _M_root->left->up = _M_root;
      }
 
      binary_tree(const binary_tree& x, allocator_type u) : _M_alloc{u}
      {
        _M_root = _M_alloc.create_root();
        _M_root->left = _M_alloc.copy_a_branch(x._M_root->left);
        _M_root->left->up = _M_root;
      }
      
      binary_tree(const binary_tree& x) : _M_alloc{x._M_alloc.resource()}
      {
        _M_root = _M_alloc.create_root();
        _M_root->left = _M_alloc.copy_a_branch(x._M_root->left);
        _M_root->left->up = _M_root;
      }
      
      binary_tree& operator=(const binary_tree& x)
      {
        if(&x != this) {
          _M_alloc.dump_a_branch(_M_root->left);
          _M_root->left = _M_alloc.copy_a_branch(x._M_root->left);
          _M_root->left->up = _M_root;
        }
        
        return *this;
      }
      
      /*   destructor   */
      ~binary_tree() {
        _M_alloc.delete_a_branch(_M_root->left);
        _M_alloc.delete_root(_M_root);
      }
      
 
      /*   initialize the head, the tree has only one element that is the head     */
      explicit binary_tree(const value_type& d, allocator_type a = nullptr)
      : _M_alloc{a} {
        _M_root = _M_alloc.create_root();
        _M_alloc.append_left(_M_root, d);
      }
      
      explicit binary_tree(value_type&& d, allocator_type a = nullptr)
      : _M_alloc{a} {
        _M_root = _M_alloc.create_root();
        _M_alloc.append_left(_M_root, std::forward<value_type>(d));
      }
      
      /*    creating the tree from the topology string   */
      explicit binary_tree(const char *, const value_type& = _Tp(), allocator_type = nullptr);
      
      template<class _InputIter>
      binary_tree(const char *, _InputIter, allocator_type = nullptr);
      
      template<class _InputIter>
      binary_tree(const char *, _InputIter, _InputIter, allocator_type = nullptr);
      
      
      /*   cursors, traversers and iterators   */
      using cursor = __tree_cursor<false>;
      using const_cursor = __tree_cursor<true>;
      
      template<class _Traversal> using traverser = __tree_traverser<_Traversal, value_type>;
      template<class _Traversal> using const_traverser = __tree_traverser<_Traversal, const value_type>;
   
      using iterator = traverser<inorder>;
      using const_iterator = const_traverser<inorder>;

      /*   root of the tree    */
      iterator root() { return traverser<inorder>{_M_root, _M_root};}
      const_iterator root() const { return const_traverser<inorder>{_M_root, _M_root};}
      const_iterator croot() const { return const_traverser<inorder>{_M_root, _M_root};}

      /*   head of the tree  */
      iterator head() { return traverser<inorder>{_M_root, _M_root->left};}
      const_iterator head() const { return const_traverser<inorder>{_M_root, _M_root->left};}
      const_iterator chead() const { return const_traverser<inorder>{_M_root, _M_root->left};}
  
      /*   default begin()/end() operators (inorder traverser)   */
      iterator begin() { return begin<inorder>();}
      iterator end() { return end<inorder>();}
 
      const_iterator begin() const { return begin<inorder>();}
      const_iterator end() const { return end<inorder>();}

      const_iterator cbegin() const { return cbegin<inorder>();}
      const_iterator cend() const { return cend<inorder>();}
      
      
      /*   non-const begin() operators   */
      template<class _Traversal>
      traverser<_Traversal> begin() {
        return traverser<_Traversal>{_M_root, __tree_traverser_impl<_Traversal, false>::begin(_M_root->left, _M_root)};
      }
  
      /*   non-const end() operators   */
      template<class _Traversal>
      traverser<_Traversal> end() {
        return traverser<_Traversal>{_M_root, _M_root};
      }

      
      /*   const begin() operators   */
      template<class _Traversal>
      const_traverser<_Traversal> begin() const {
        return const_traverser<_Traversal>{_M_root, __tree_traverser_impl<_Traversal, true>::begin(_M_root->left, _M_root)};
      }
      
      template<class _Traversal>
      const_traverser<_Traversal> cbegin() const {
        return const_traverser<_Traversal>{_M_root, __tree_traverser_impl<_Traversal, true>::begin(_M_root->left, _M_root)};
      }
      
      /*   const end() operators   */
      template<class _Traversal>
      const_traverser<_Traversal> end() const {
        return const_traverser<_Traversal>{_M_root, _M_root};
      }
      
      template<class _Traversal>
      const_traverser<_Traversal> cend() const {
        return const_traverser<_Traversal>{_M_root, _M_root};
      }
      
      /*   reverse traversers   */
      template<class _Traversal> using reverse_traverser = std::reverse_iterator<traverser<_Traversal>>;
      template<class _Traversal> using const_reverse_traverser = std::reverse_iterator<const_traverser<_Traversal>>;
      
      using reverse_iterator = reverse_traverser<inorder>;
      using const_reverse_iterator = const_reverse_traverser<inorder>;

      
      /*   default begin()/end() operators (inorder traverser)   */
      reverse_iterator rbegin() { return rbegin<inorder>();}
      reverse_iterator rend() { return rend<inorder>();}
      
      const_reverse_iterator rbegin() const { return rbegin<inorder>();}
      const_reverse_iterator rend() const { return rend<inorder>();}
      
      const_reverse_iterator crbegin() const { return crbegin<inorder>();}
      const_reverse_iterator crend() const { return crend<inorder>();}
      
      
      /*   non-const rbegin() operators   */
      template<class _Traversal>
      reverse_traverser<_Traversal> rbegin() {
        return reverse_traverser<_Traversal>{end<_Traversal>()};
      }
      
      /*   non-const rend() operators   */
      template<class _Traversal>
      reverse_traverser<_Traversal> rend() {
        return reverse_traverser<_Traversal>{begin<_Traversal>()};
      }
      
    
      /*   const rbegin() operators   */
      template<class _Traversal>
      const_reverse_traverser<_Traversal> rbegin() const {
        return const_reverse_traverser<_Traversal>{end<_Traversal>()};
      }
      
      template<class _Traversal>
      const_reverse_traverser<_Traversal> crbegin() const {
        return const_reverse_traverser<_Traversal>{cend<_Traversal>()};
      }
      
      /*   const rend() operators   */
      template<class _Traversal>
      const_reverse_traverser<_Traversal> rend() const {
        return const_reverse_traverser<_Traversal>{begin<_Traversal>()};
      }
      
      template<class _Traversal>
      const_reverse_traverser<_Traversal> crend() const {
        return const_reverse_traverser<_Traversal>{cbegin<_Traversal>()};
      }
      
      
      /*    returns a pair of iterator as range   */
      iterator_range<iterator> range() { return range<inorder>();}
      iterator_range<const_iterator> range() const{ return range<inorder>();}
      iterator_range<const_iterator> crange() const { return crange<inorder>();}
      
      template<class _Traversal>
      iterator_range<traverser<_Traversal>>
      range() { return {begin<_Traversal>(), end<_Traversal>()};}

      template<class _Traversal>
      iterator_range<const_traverser<_Traversal>>
      range() const { return {begin<_Traversal>(), end<_Traversal>()};}
      
      template<class _Traversal>
      iterator_range<const_traverser<_Traversal>>
      crange() const { return {cbegin<_Traversal>(), cend<_Traversal>()};}
      
      
      /*    reverese rangees    */
      iterator_range<reverse_iterator> rrange() { return rrange<inorder>();}
      iterator_range<const_reverse_iterator> rrange() const{ return rrange<inorder>();}
      iterator_range<const_reverse_iterator> crrange() const { return crrange<inorder>();}

      template<class _Traversal>
      iterator_range<reverse_traverser<_Traversal>>
      rrange() { return {rbegin<_Traversal>(), rend<_Traversal>()};}

      template<class _Traversal>
      iterator_range<const_reverse_traverser<_Traversal>>
      rrange() const { return {rbegin<_Traversal>(), rend<_Traversal>()};}
      
      template<class _Traversal>
      iterator_range<const_reverse_traverser<_Traversal>>
      crrange() const { return {crbegin<_Traversal>(), crend<_Traversal>()};}
      

      /*   append a node to the left if it is a free slot  */
      cursor append_left(cursor iter) {
        return {_M_root, _M_alloc.append_left(iter)};
      }

      cursor append_left(cursor iter, value_type&& v) {
        return {_M_root, _M_alloc.append_left(iter, std::forward<value_type>(v))};
      }
      
      cursor append_left(cursor iter, const value_type& v) {
        return {_M_root, _M_alloc.append_left(iter, v)};
      }
      
      /*   append a node to the right if it is a free slot  */
      cursor append_right(cursor iter) {
        return {_M_root, _M_alloc.append_right(iter)};
      }
      
      cursor append_right(cursor iter, value_type&& v) {
        return {_M_root, _M_alloc.append_right(iter, std::forward<value_type>(v))};
      }
      
      cursor append_right(cursor iter, const value_type& v) {
        return {_M_root, _M_alloc.append_right(iter, v)};
      }
     
      /*    appending a node by emplacing the value    */
      template<typename ..._Args>
      cursor emplace_left(cursor iter, _Args&& ...args) {
        return {_M_root, _M_alloc.emplace_left(iter, std::forward<_Args>(args)...)};
      }
      
      template<typename ..._Args>
      cursor emplace_right(cursor iter, _Args&& ...args) {
        return {_M_root, _M_alloc.emplace_right(iter, std::forward<_Args>(args)...)};
      }
   
      
      /*   append a tree to the left if it is a free slot   */
      cursor append_left(cursor pos, binary_tree&& d)
      {
        __tree_node_link *n = pos;
        
        if(n->left == nullptr) {
          n->left = _M_bare_tree(std::move(d));
          n->left->up = n;
          return {_M_root, n->left};
        }

        return pos;
      }
      
      cursor append_left(cursor pos, const binary_tree& d)
      {
        __tree_node_link *n = pos;
        
        if(n->left == nullptr) {
          n->left = _M_bare_tree(d);
          n->left->up = n;
          return {_M_root, n->left};
        }
        
        return pos;
      }
      
      
      /*   append a tree to the right if it is a free slot   */
      cursor append_right(cursor pos, binary_tree&& d)
      {
        __tree_node_link *n = pos;
        
        if(n->right == nullptr) {
          n->right = _M_bare_tree(std::move(d));
          n->right->up = n;
          return {_M_root, n->right};
        }
        
        return pos;
      }
      
      cursor append_right(cursor pos, const binary_tree& d)
      {
        __tree_node_link *n = pos;
        
        if(n->right == nullptr) {
          n->right = _M_bare_tree(d);
          n->right->up = n;
          return {_M_root, n->right};
        }
        
        return pos;
      }
      
      
      /*   inserting  a node with left/right leg    */
      cursor insert_left(cursor iter) {
        return {_M_root, _M_alloc.insert_left(iter)};
      }
      
      cursor insert_right(cursor iter) {
        return {_M_root, _M_alloc.insert_right(iter)};
      }
      
     cursor insert_left(cursor iter, value_type&& v) {
        return {_M_root, _M_alloc.insert_left(iter, std::forward<value_type>(v))};
      }
      
      cursor insert_right(cursor iter, value_type&& v) {
        return {_M_root, _M_alloc.insert_right(iter, std::forward<value_type>(v))};
      }
 
      cursor insert_left(cursor iter, const value_type& v) {
        return {_M_root, _M_alloc.insert_left(iter, v)};
      }
      
      cursor insert_right(cursor iter, const value_type& v) {
        return {_M_root, _M_alloc.insert_right(iter, v)};
      }
  
      /*   inserting a node by emplacing the new element   */
      template<typename ..._Args>
      cursor emplance_insert_left(cursor iter, _Args&& ...args) {
        return {_M_root, _M_alloc.emplance_insert_left(iter, std::forward<_Args>(args)...)};
      }
      
      template<typename ..._Args>
      cursor emplace_insert_right(cursor iter, _Args&& ...args) {
        return {_M_root, _M_alloc.emplace_insert_right(iter, std::forward<_Args>(args)...)};
      }
      
      /*   erase the subtree started with the given node   */
      void erase_branch(cursor pos) {
        _M_alloc.dump_a_branch(pos);
      }
      
      /*   split the tree at the given position  */
      binary_tree split(cursor pos)
      {
        /*  nothing to do */
        if(pos == _M_root) return binary_tree{_M_alloc.resource()};
        
        /*  position is the head   */
        if(_M_root->left == pos) {
          binary_tree ret{_M_alloc.resource()};
          this->swap(ret);
          return ret;
        }
        
        /*   split somewhere in the middle   */
        __tree_node_link *c = pos;
        
        /*   separeting form the rest of the tree   */
        if(c->up->left == c) c->up->left = nullptr;
        else c->up->right = nullptr;
        
        /*   returning the subtree  */
        binary_tree ret{_M_alloc.resource()};
       
        ret._M_root->left = c;
        ret._M_root->left->up = ret._M_root;
        
        return ret;
      }
      
      /*   swapping trees (do not swap allocators) */
      void swap(binary_tree& a) { std::swap(_M_root, a._M_root);}
      
      /*   dump all the nodes   */
      void clear() { _M_alloc.dump_a_branch(_M_root->left);}
      
      /*   delete all the junk nodes   */
      void shrink_to_fit() { _M_alloc.burn_junk();}
      
      /*   structural information  */
      size_type size() const { return std::distance(begin(), end()); }
      bool empty() const { return _M_root->left == nullptr;}
     
      /*   returns a pointer to the resource object   */
      pmr::memory_resource * resource() const { return _M_alloc.resource();}
      pmr::memory_resource * get_allocator() const { return resource();}

      /*   topology   */
      std::string topology() const {
        return container::topology(_M_root->left);
      }

    private:
      /*   the end of the tree   */
      __tree_node_link *_M_root;
      
      /*   node allocator   */
      node_allocator _M_alloc;

      /*   bare tree:  move constructor  */
      __tree_node_link * _M_bare_tree(binary_tree&& x)
      {
        __tree_node_link *head;
        
        if(_M_alloc.resource()->is_equal(_M_alloc.resource())) {
          head = x._M_root->left;
          x._M_root->left = head->up = nullptr;
        }
        /*   move the node's data one-by-one   */
        else {
          head = _M_alloc.move_a_branch(x._M_root->left);
          head->up = nullptr;
        }
        
        return head;
      }
      
      /*   bare tree:  copy constructor  */
      __tree_node_link * _M_bare_tree(const binary_tree& x) {
        return _M_alloc.copy_a_branch(x._M_root->left);
      }
    };
    
    
    /*    creating the tree from the topology string   */
    template<typename _Tp, bool _Recycle>
    binary_tree<_Tp, _Recycle>::binary_tree(const char *s, const value_type& v, allocator_type a)
    : _M_alloc{a}
    {
      _M_root = _M_alloc.create_root();
      __tree_node_link *c = _M_root;
      
      /*   building the tree    */
      for(;*s != '\0' && *s != 'N'; ++s) {
        switch(*s) {
          case 'H' :
          case 'L' : c = _M_alloc.append_left(c, v); break;
          case 'R' : c = _M_alloc.append_right(c, v); break;
          case 'u' : c = c->up; break;
          default: continue;
        }
        
        /*   wrong string can lead to trouble...   */
        if(c == _M_root || c == _M_root->up) break;
      }
    }
    
    
    template<typename _Tp, bool _Recycle> template<class _InputIter>
    binary_tree<_Tp, _Recycle>::binary_tree(const char *s, _InputIter v, allocator_type a)
    : _M_alloc{a}
    {
      _M_root = _M_alloc.create_root();
      __tree_node_link *c = _M_root;
      
      /*   building the tree    */
      for(;*s != '\0' && *s != 'N'; ++s) {
        switch(*s) {
          case 'H' :
          case 'L' : c = _M_alloc.append_left(c, *v); ++v; break;
          case 'R' : c = _M_alloc.append_right(c, *v); ++v; break;
          case 'u' : c = c->up; break;
          default: continue;
        }
        
        /*   wrong string can lead to trouble...   */
        if(c == _M_root || c == _M_root->up) break;
      }
    }
    
    
    template<typename _Tp, bool _Recycle> template<class _InputIter>
    binary_tree<_Tp, _Recycle>::binary_tree(const char *s, _InputIter v, _InputIter vend, allocator_type a)
    : _M_alloc{a}
    {
      _M_root = _M_alloc.create_root();
      __tree_node_link *c = _M_root;
      
      /*   building the tree    */
      for(;*s != '\0' && *s != 'N'; ++s) {
        switch(*s) {
          case 'H' :
          case 'L' : c = _M_alloc.append_left(c, v == vend ? _Tp() : *v++); break;
          case 'R' : c = _M_alloc.append_right(c, v == vend ? _Tp() : *v++); break;
          case 'u' : c = c->up; break;
          default: continue;
        }
        
        /*   wrong string can lead to trouble...   */
        if(c == _M_root || c == _M_root->up) break;
      }
    }
    
    
    template<typename _Tp, bool _Recycle> inline
    void swap(binary_tree<_Tp, _Recycle>& a, binary_tree<_Tp, _Recycle>& b) {
      a.swap(b);
    }
    

    
    /*    subtree iterators   */
    template<class _Traversal, typename _Tp, bool _Recycle> inline
    typename binary_tree<_Tp, _Recycle>::template traverser<_Traversal>
    begin(const typename binary_tree<_Tp, _Recycle>::template traverser<_Traversal>& subtree) {
      __tree_node_link *s = subtree;
      return typename binary_tree<_Tp, _Recycle>::template traverser<_Traversal>{s->up, __tree_traverser_impl<_Traversal, false>::begin(s, s->up)};
    }
    
    template<class _Traversal, typename _Tp, bool _Recycle> inline
    typename binary_tree<_Tp, _Recycle>::template traverser<_Traversal>
    end(const typename binary_tree<_Tp, _Recycle>::template traverser<_Traversal>& subtree) {
      __tree_node_link *s = subtree;
      return typename binary_tree<_Tp, _Recycle>::template traverser<_Traversal>{s->up, s->up};
    }
    
    template<class _Traversal, typename _Tp, bool _Recycle> inline
    typename binary_tree<_Tp, _Recycle>::template const_traverser<_Traversal>
    begin(const typename binary_tree<_Tp, _Recycle>::template const_traverser<_Traversal>& subtree) {
      __tree_node_link *s = subtree;
      return typename binary_tree<_Tp, _Recycle>::template const_traverser<_Traversal>{s->up, __tree_traverser_impl<_Traversal, true>::begin(s, s->up)};
    }
    
    template<class _Traversal, typename _Tp, bool _Recycle> inline
    typename binary_tree<_Tp, _Recycle>::template const_traverser<_Traversal>
    end(const typename binary_tree<_Tp, _Recycle>::template const_traverser<_Traversal>& subtree) {
      __tree_node_link *s = subtree;
      return typename binary_tree<_Tp, _Recycle>::template const_traverser<_Traversal>{s->up, s->up};
    }
    
    template<class _Traversal, typename _Tp, bool _Recycle> inline
    typename binary_tree<_Tp, _Recycle>::template const_traverser<_Traversal>
    cbegin(const typename binary_tree<_Tp, _Recycle>::template traverser<_Traversal>& subtree) {
      __tree_node_link *s = subtree;
      return typename binary_tree<_Tp, _Recycle>::template const_traverser<_Traversal>{s->up, __tree_traverser_impl<_Traversal, true>::begin(s, s->up)};
    }
    
    template<class _Traversal, typename _Tp, bool _Recycle> inline
    typename binary_tree<_Tp, _Recycle>::template const_traverser<_Traversal>
    cend(const typename binary_tree<_Tp, _Recycle>::template traverser<_Traversal>& subtree) {
      __tree_node_link *s = subtree;
      return typename binary_tree<_Tp, _Recycle>::template const_traverser<_Traversal>{s->up, s->up};
    }
    
    template<class _Traversal, typename _Tp, bool _Recycle> inline
    typename binary_tree<_Tp, _Recycle>::template const_traverser<_Traversal>
    cbegin(const typename binary_tree<_Tp, _Recycle>::template const_traverser<_Traversal>& subtree) {
      __tree_node_link *s = subtree;
      return typename binary_tree<_Tp, _Recycle>::template const_traverser<_Traversal>{s->up, __tree_traverser_impl<_Traversal, true>::begin(s, s->up)};
    }
    
    template<class _Traversal, typename _Tp, bool _Recycle> inline
    typename binary_tree<_Tp, _Recycle>::template const_traverser<_Traversal>
    cend(const typename binary_tree<_Tp, _Recycle>::template const_traverser<_Traversal>& subtree) {
      __tree_node_link *s = subtree;
      return typename binary_tree<_Tp, _Recycle>::template const_traverser<_Traversal>{s->up, s->up};
    }
    
    
    /*    subtree reverse iterators   */
    template<class _Traversal, typename _Tp, bool _Recycle> inline
    typename binary_tree<_Tp, _Recycle>::template reverse_traverser<_Traversal>
    rbegin(const typename binary_tree<_Tp, _Recycle>::template traverser<_Traversal>& subtree) {
      return typename binary_tree<_Tp, _Recycle>::template reverse_traverser<_Traversal>{end<_Traversal>(subtree)};
    }
    
    template<class _Traversal, typename _Tp, bool _Recycle> inline
    typename binary_tree<_Tp, _Recycle>::template reverse_traverser<_Traversal>
    rend(const typename binary_tree<_Tp, _Recycle>::template traverser<_Traversal>& subtree) {
      return typename binary_tree<_Tp, _Recycle>::template reverse_traverser<_Traversal>{begin<_Traversal>(subtree)};
    }
    
    template<class _Traversal, typename _Tp, bool _Recycle> inline
    typename binary_tree<_Tp, _Recycle>::template const_reverse_traverser<_Traversal>
    rbegin(const typename binary_tree<_Tp, _Recycle>::template const_traverser<_Traversal>& subtree) {
      return typename binary_tree<_Tp, _Recycle>::template const_reverse_traverser<_Traversal>{end<_Traversal>(subtree)};
    }
    
    template<class _Traversal, typename _Tp, bool _Recycle> inline
    typename binary_tree<_Tp, _Recycle>::template const_reverse_traverser<_Traversal>
    rend(const typename binary_tree<_Tp, _Recycle>::template const_traverser<_Traversal>& subtree) {
      return typename binary_tree<_Tp, _Recycle>::template const_reverse_traverser<_Traversal>{begin<_Traversal>(subtree)};
    }
    
    template<class _Traversal, typename _Tp, bool _Recycle> inline
    typename binary_tree<_Tp, _Recycle>::template const_reverse_traverser<_Traversal>
    crbegin(const typename binary_tree<_Tp, _Recycle>::template traverser<_Traversal>& subtree) {
      return typename binary_tree<_Tp, _Recycle>::template const_reverse_traverser<_Traversal>{cend<_Traversal>(subtree)};
    }
    
    template<class _Traversal, typename _Tp, bool _Recycle> inline
    typename binary_tree<_Tp, _Recycle>::template const_reverse_traverser<_Traversal>
    crend(const typename binary_tree<_Tp, _Recycle>::template traverser<_Traversal>& subtree) {
      return typename binary_tree<_Tp, _Recycle>::template const_reverse_traverser<_Traversal>{cbegin<_Traversal>(subtree)};
    }
    
    template<class _Traversal, typename _Tp, bool _Recycle> inline
    typename binary_tree<_Tp, _Recycle>::template const_reverse_traverser<_Traversal>
    crbegin(const typename binary_tree<_Tp, _Recycle>::template const_traverser<_Traversal>& subtree) {
      return typename binary_tree<_Tp, _Recycle>::template const_reverse_traverser<_Traversal>{cend<_Traversal>(subtree)};
    }
    
    template<class _Traversal, typename _Tp, bool _Recycle> inline
    typename binary_tree<_Tp, _Recycle>::template const_reverse_traverser<_Traversal>
    crend(const typename binary_tree<_Tp, _Recycle>::template const_traverser<_Traversal>& subtree) {
      return typename binary_tree<_Tp, _Recycle>::template const_reverse_traverser<_Traversal>{cbegin<_Traversal>(subtree)};
    }

    
    /*    subtree ranges    */
    template<class _Traversal, typename _Tp, bool _Recycle> inline
    iterator_range<typename binary_tree<_Tp, _Recycle>::template traverser<_Traversal>>
    branch(const typename binary_tree<_Tp, _Recycle>::template traverser<_Traversal>& s) {
      return {begin(s), end(s)};
    }

    template<class _Traversal, typename _Tp, bool _Recycle> inline
    iterator_range<typename binary_tree<_Tp, _Recycle>::template const_traverser<_Traversal>>
    branch(const typename binary_tree<_Tp, _Recycle>::template const_traverser<_Traversal>& s) {
      return {begin(s), end(s)};
    }
    
    template<class _Traversal, typename _Tp, bool _Recycle> inline
    iterator_range<typename binary_tree<_Tp, _Recycle>::template const_traverser<_Traversal>>
    cbranch(const typename binary_tree<_Tp, _Recycle>::template traverser<_Traversal>& s) {
      return {cbegin(s), cend(s)};
    }
    
    
    /*    subtree ranges  (reverse iterator)  */
    template<class _Traversal, typename _Tp, bool _Recycle> inline
    iterator_range<typename binary_tree<_Tp, _Recycle>::template reverse_traverser<_Traversal>>
    rbranch(const typename binary_tree<_Tp, _Recycle>::template traverser<_Traversal>& s) {
      return {rbegin(s), rend(s)};
    }
    
    template<class _Traversal, typename _Tp, bool _Recycle> inline
    iterator_range<typename binary_tree<_Tp, _Recycle>::template const_reverse_traverser<_Traversal>>
    rbranch(const typename binary_tree<_Tp, _Recycle>::template const_traverser<_Traversal>& s) {
      return {rbegin(s), rend(s)};
    }
    
    template<class _Traversal, typename _Tp, bool _Recycle> inline
    iterator_range<typename binary_tree<_Tp, _Recycle>::template const_reverse_traverser<_Traversal>>
    crbranch(const typename binary_tree<_Tp, _Recycle>::template traverser<_Traversal>& s) {
      return {crbegin(s), crend(s)};
    }
    
    
  }   //   namespace containers
}

#endif
