/*  Copyright (C) 2002-2017 Zoltán Nagy
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __duct_lorentzvector_h__
#define __duct_lorentzvector_h__ 1


/*   std includes   */


/*  project includes   */
#include <bits/duct-threevector.h>



namespace duct {
  
  
  /*  Template class for representing lorentz vectors.
   *  (object with single lorentz index)
   */
  template<typename _Tp>
  class lorentzvector : public threevector<_Tp>
  {
    /*   private types   */
    using _Base = threevector<_Tp>;
    
  public:
    using value_type = typename _Base::value_type;
    using threevector_type = _Base;
    
    
    /*   default constructor  */
    lorentzvector() {
      _Base::_M_c[0] = value_type{};
    }
    
    /*   constructors   */
    lorentzvector(value_type E, value_type px, value_type py, value_type pz)
    {
      _Base::_M_c[0] = std::move(E);
      _Base::_M_c[1] = std::move(px);
      _Base::_M_c[2] = std::move(py);
      _Base::_M_c[3] = std::move(pz);
    }
    
    template<typename _Xp>
    lorentzvector(value_type E, const threevector<_Xp>& v)
    : _Base{v} { _Base::_M_c[0] = std::move(E);}
    
    
    /*   move   */
    lorentzvector(lorentzvector&&) = default;
    lorentzvector& operator=(lorentzvector&&) = default;
    
    
    /*   copy   */
    lorentzvector(const lorentzvector& p)
    : _Base{p[1], p[2], p[3]} { _Base::_M_c[0] = p[0];}
    
    
    template<typename _Xp>
    explicit lorentzvector(const lorentzvector<_Xp>& p) {
      _Base::_M_copy(p[0], p);
    }
    
    lorentzvector& operator=(const lorentzvector& p) {
      if(this != &p) _Base::_M_copy(p[0], p);
      return *this;
    }
    
    template<typename _Xp>
    lorentzvector& operator=(const lorentzvector<_Xp>& p) {
      _Base::_M_copy(p[0], p);
      return *this;
    }
    
    
    /*   computed assinments   */
    template<typename _Xp>
    lorentzvector& operator+=(const lorentzvector<_Xp>& p) {
      _Base::_M_c[0] += p[0];
      _Base::operator+=(p);
      return *this;
    }
    
    template<typename _Xp>
    lorentzvector& operator-=(const lorentzvector<_Xp>& p) {
      _Base::_M_c[0] -= p[0];
      _Base::operator-=(p);
      return *this;
    }
    
    template<typename _Xp>
    lorentzvector& operator*=(const _Xp& x) {
      _Base::_M_c[0] *= x;
      _Base::operator*=(x);
      return *this;
    }
    
    template<typename _Xp>
    lorentzvector& operator/=(const _Xp& x) {
      _Base::_M_c[0] /= x;
      _Base::operator/=(x);
      return *this;
    }
    
    
    /*   light cone coordinates   */
    value_type plus () const { return _Base::_M_c[0] + _Base::_M_c[3];}
    value_type minus() const { return _Base::_M_c[0] - _Base::_M_c[3];}
    
    
    /*    rapidities    */
    value_type rapidity() const requires RealNumbers<value_type> {
      return value_type{0.5}*std::log(plus()/minus());
    }
    
    value_type pseudo_rapidity() const requires RealNumbers<value_type> {
      return -std::log(std::tan(value_type{0.5}*_Base::theta()));
    }
    
    /*   invariant mass square   */
    value_type mag2() const {
      return _Base::_M_c[0]*_Base::_M_c[0] - _Base::mag2();
    }
    
    /*   invariant mass    */
    value_type mag() const requires RealNumbers<value_type> {
      value_type p =  mag2();
      return p >= 0.0 ? std::sqrt(p) : -std::sqrt(-p);
    }
    
    /*   tranverse momentum   */
    template<RealNumbers _Xp, RealNumbers _Yp>
    lorentzvector transverse(const lorentzvector<_Xp>&, const lorentzvector<_Yp>&) const;
    
    /*   boost vector    */
    threevector_type boostVector() const requires RealNumbers<value_type> {
      return _Base{*this} /= _Base::_M_c[0];
    }
    
    /*   Lorentz boost   */
    void boost(RealNumbers, RealNumbers, RealNumbers);
    
    void boost(const threevector<RealNumbers>& a) {
      boost(a[1], a[2], a[3]);
    }
  };
  
  
  /*   Lorentz boost   */
  template<typename _Tp>
  void lorentzvector<_Tp>::boost(RealNumbers bx, RealNumbers by, RealNumbers bz)
  {
    auto b2 = bx*bx + by*by + bz*bz;
    auto gamma = 1.0/std::sqrt(1.0 - b2);
    auto gamma2 = (b2 > 0.0 ? (gamma - 1.0)/b2 : 0.0);
    
    auto bp = bx*_Base::_M_c[1] + by*_Base::_M_c[2] + bz*_Base::_M_c[3];
    
    _Base::_M_c[1] += (gamma2*bx)*bp + (gamma*bx)*_Base::_M_c[0];
    _Base::_M_c[2] += (gamma2*by)*bp + (gamma*by)*_Base::_M_c[0];
    _Base::_M_c[3] += (gamma2*bz)*bp + (gamma*bz)*_Base::_M_c[0];
    _Base::_M_c[0]  = gamma*(_Base::_M_c[0] + bp);
  }
  
  
  /*   Transverse projection    */
  template<typename _Tp>
  template<RealNumbers _Xp, RealNumbers _Yp>
  lorentzvector<_Tp> lorentzvector<_Tp>::
  transverse(const lorentzvector<_Xp>& na, const lorentzvector<_Yp>& nb) const
  {
    auto sab = na*nb;
    auto na2 = na.mag2();
    auto nb2 = nb.mag2();
    
    const auto& p = *this;
    auto ra = na - nb*(na2/sab);
    auto rb = nb - na*(nb2/sab);
    
    auto retval = na*dot(rb, p);
    
    retval += nb*dot(ra, p);
    retval *= sab/(sab*sab-na2*nb2);
    retval -= p;
    
    return retval;
  }
  
  
  
  /*   lightcone variables   */
  inline auto plus(const lorentzvector<auto>& p) {
    return p.plus();
  }
  
  inline auto minus(const lorentzvector<auto>& p) {
    return p.minus();
  }
  
  
  /*    rapidities    */
  inline auto rapidity(const lorentzvector<RealNumbers>& p) {
    return p.rapidity();
  }
  
  inline auto pseudo_rapidity(const lorentzvector<RealNumbers>& p) {
    return p.pseudo_rapidity();
  }
  
  /*   invariant mass square   */
  inline auto mag2(const lorentzvector<auto>& p) {
    return p.mag2();
  }
  
  /*   invariant mass   */
  inline auto mag(const lorentzvector<RealNumbers>& p) {
    return p.mag();
  }
  
  /*   boost vector   */
  inline auto boostVector(const lorentzvector<RealNumbers>& p) {
    return p.boostVector();
  }
  
  
  /*   unary operators    */
  template<typename _Tp> inline
  constexpr const lorentzvector<_Tp>& operator+(const lorentzvector<_Tp>& p) {
    return p;
  }
  
  template<typename _Tp>
  inline lorentzvector<_Tp> operator-(const lorentzvector<_Tp>& p) {
    return {-p[0], -p[1], -p[2], -p[3]};
  }
  
  
  /*   arithmetic binary operators   */
  template<typename _Tp, typename _Xp>
  inline auto operator+(const lorentzvector<_Tp>& a, const lorentzvector<_Xp>& b)
  -> lorentzvector<decltype(std::declval<_Tp>()+std::declval<_Xp>())>
  {
    return {a[0]+b[0], a[1]+b[1], a[2]+b[2], a[3]+b[3]};
  }
  
  template<typename _Tp, typename _Xp>
  inline auto operator-(const lorentzvector<_Tp>& a, const lorentzvector<_Xp>& b)
  -> lorentzvector<decltype(std::declval<_Tp>()-std::declval<_Xp>())>
  {
    return {a[0]-b[0], a[1]-b[1], a[2]-b[2], a[3]-b[3]};
  }
  
  
  template<typename _Tp, typename _Xp>
  inline auto operator*(const lorentzvector<_Tp>& a, const _Xp& b)
  -> lorentzvector<decltype(std::declval<_Tp>()*b)>
  {
    return {a[0]*b, a[1]*b, a[2]*b, a[3]*b};
  }
  
  template<typename _Tp, typename _Xp>
  inline auto operator*(const _Xp& b, const lorentzvector<_Tp>& a)
  -> lorentzvector<decltype(std::declval<_Tp>()*b)>
  {
    return {a[0]*b, a[1]*b, a[2]*b, a[3]*b};
  }
  
  template<typename _Tp, typename _Xp>
  inline auto operator/(const lorentzvector<_Tp>& a, const _Xp& b)
  -> lorentzvector<decltype(std::declval<_Tp>()/b)>
  {
    return {a[0]/b, a[1]/b, a[2]/b, a[3]/b};
  }
  
  
  /*    comparisons    */
  template<typename _Tp, typename _Xp>
  inline bool operator==(const lorentzvector<_Tp>& a, const lorentzvector<_Xp>& b) {
    return a[0] == b[0] && a[1] == b[1] && a[2] == b[2] && a[3] == b[3];
  }
  
  
  template<typename _Tp, typename _Xp>
  inline bool operator!=(const lorentzvector<_Tp>& a, const lorentzvector<_Xp>& b) {
    return a[0] != b[0] || a[1] != b[1] || a[2] != b[2] || a[3] != b[3];
  }
  
  
  /*    dot products   */
  template<typename _Tp, typename _Xp>
  inline auto operator*(const lorentzvector<_Tp>& a, const lorentzvector<_Xp>& b)
  -> decltype(std::declval<_Tp>()*std::declval<_Xp>())
  {
    return a[0]*b[0] - a[1]*b[1] - a[2]*b[2] - a[3]*b[3];
  }
  
  
  template<typename _Tp, typename _Xp>
  inline auto dot(const lorentzvector<_Tp>& a, const lorentzvector<_Xp>& b)
  -> decltype(std::declval<_Tp>()*std::declval<_Xp>())
  {
    return a[0]*b[0] - a[1]*b[1] - a[2]*b[2] - a[3]*b[3];
  }
  
  
  template<RealNumbers _Tp, RealNumbers _Xp>
  auto scalar_mp(lorentzvector<_Tp> a, lorentzvector<_Xp> b)
  {
    using _ComplexD = std::complex<std::common_type_t<_Tp, _Xp>>;
    
    constexpr _ComplexD I{0.0,1.0};
    
    _ComplexD ea{1.0}, eb{1.0}, ph{1.0};
    auto pa = a.perp();
    auto pb = b.perp();
    
    if(a[0] < 0.0) a *= -1.0, ph *= I;
    if(b[0] < 0.0) b *= -1.0, ph *= I;
    
    if(pa != 0.0) ea = _ComplexD{a[1], a[2]}/pa;
    if(pb != 0.0) eb = _ComplexD{b[1], b[2]}/pb;
    
    auto mp = a.minus()*b.plus();
    auto pm = a.plus()*b.minus();
    
    mp = (mp <= 0.0 ? 0.0 : mp);
    pm = (pm <= 0.0 ? 0.0 : pm);
    
    return ph*(std::sqrt(mp)*ea - std::sqrt(pm)*eb);
  }
  

  /*    Complex operators   */
  template<RealNumbers _Tp>
  inline lorentzvector<_Tp> real(const lorentzvector<std::complex<_Tp>>& a) {
    return {std::real(a[0]), std::real(a[1]), std::real(a[2]), std::real(a[3])};
  }
  
  template<RealNumbers _Tp>
  inline lorentzvector<_Tp> imag(const lorentzvector<std::complex<_Tp>>& a) {
    return {std::imag(a[0]), std::imag(a[1]), std::imag(a[2]), std::imag(a[3])};
  }
  
  template<RealNumbers _Tp>
  inline lorentzvector<std::complex<_Tp>> conj(const lorentzvector<std::complex<_Tp>>& a) {
    return {std::conj(a[0]), std::conj(a[1]), std::conj(a[2]), std::conj(a[3])};
  }
  
  
  
  /*    I/O operations    */
  inline std::ostream& operator<<(std::ostream& os, const lorentzvector<auto>& q) {
    return os<<"["<<q[0]<<", ("<<q[1]<<", "<<q[2]<<", "<<q[3]<<")]";
  }
  
}  //   namespace duct

#endif
