/*  Copyright (C) 2002-2013 Zoltán Nagy
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __deductor_hep_dist__
#define __deductor_hep_dist__ 1


//   Standard includes
#include <list>
#include <vector>
#include <iostream>

//   deductor includes
#include <bits/hep-sample_statistic.h>
#include <cxx/binary-io.h>


namespace duct {
  
 
  /*   Some predefined goodies to implement 
   *   simple distributions and histograms.
   */
  namespace distpoint {
    
    
    /*   simple tag tto define zero dimensional histograms   */
    struct nulld {};
  
    
    //    a simple point type for the 1D histograms
    struct hist1d {
      double xmin, xmax;
    };
    
    inline std::ostream& write(std::ostream& os, const hist1d& p) {
      return duct::write(os, p.xmin, p.xmax);
    }
    
    inline std::istream& read(std::istream& is, hist1d& p) {
      return duct::read(is, p.xmin, p.xmax);
    }

    
    
    //    a simple point type for the 2D histograms
    struct hist2d {
      double xmin, xmax;
      double ymin, ymax;
    };
    
    inline std::ostream& write(std::ostream& os, const hist2d& p) {
      return duct::write(os, p.xmin, p.xmax, p.ymin, p.ymax);
    }
    
    inline std::istream& read(std::istream& is, hist2d& p) {
      return duct::read(is, p.xmin, p.xmax, p.ymin, p.ymax);
    }
    
    
   
    
    /*   step function  */
    struct step
    {
      double operator()(double y, double x) const  {
        return x > y ? 1.0 : 0.0;
      }

      double operator()(const hist1d& y, double x) const {
        return x > y.xmin ? 1.0 : 0.0;
      }
      
      double operator()(const hist2d& h, double x, double y) const
      {
        if(x <= h.xmin) return 0.0;
        if(y <= h.ymin) return 0.0;
        return 1.0;
      }
    };
    
    
    /*   simple apprixmation of the Dirac delta function  */
    struct dirac
    {
      double operator()(const hist1d& y, double x) const {
        return y.xmin <= x && x < y.xmax ? 1.0/(y.xmax-y.xmin) : 0.0;
      }
      
      double operator()(const hist2d& h, double x, double y) const
      {
        if(!(h.xmin <= x && x < h.xmax)) return 0.0;
        if(!(h.ymin <= y && y < h.ymax)) return 0.0;
        return 1.0/(h.xmax-h.xmin)/(h.ymax-h.ymin);
      }
    };
    
    struct stepdirac
    {
      double operator()(const hist2d& h, double x, double y) const
      {
        if(x <= h.xmin) return 0.0;
        if(!(h.ymin <= y && y < h.ymax)) return 0.0;
        return 1.0/(h.ymax-h.ymin);
      }
    };
    
    struct diracstep
    {
      double operator()(const hist2d& h, double x, double y) const
      {
        if(!(h.xmin <= x && x < h.xmax)) return 0.0;
        if(y <= h.ymin) return 0.0;
        return 1.0/(h.xmax-h.xmin);
      }
    };
    
    
    
    /*   simple class to setup distpoints   */
    template<typename _Tp>
    struct spacing;
    
    
    template<typename _Float>
    requires std::is_floating_point<_Float>::value
    struct spacing<_Float>
    {
      /*   linear spacing   */
      static  std::vector<_Float> linear(unsigned int n, _Float xmin, _Float xmax)
      {
        std::vector<_Float> p(n+1, xmin);
        float dx = (xmax-xmin)/n;
        
        for(unsigned i = 1; i <= n; i++)
          p[i] += i*dx;
        
        return  p;
      }
      
      /*   logarithmic spacing   */
      static std::vector<_Float> logarithmic(unsigned int n, _Float xmin, _Float xmax)
      {
        std::vector<_Float> p(n+1, xmin);
        
        for(unsigned i = 1; i <= n; i++)
          p[i] *= std::pow(xmax/xmin, static_cast<_Float>(i)/n);
        
        return p;
      }
    };
    
    
    
    template<>
    struct spacing<hist1d> {
      /*   linear spacing   */
      static std::vector<hist1d> linear(unsigned, double, double);
      
      /*   logarithmic spacing   */
      static std::vector<hist1d> logarithmic(unsigned, double, double);

      /*   manual spacing   */
      static std::vector<hist1d> manual(std::initializer_list<double>);
    };
    
    
    template<>
    struct spacing<hist2d> {
      /*   linear spacing   */
      static std::vector<hist2d>
      linear(unsigned, double, double, unsigned, double, double);
      
      /*   logarithmic spacing   */
      static std::vector<hist2d>
      logarithmic(unsigned, double, double, unsigned, double, double);
      
      /*   logarithmic-linear spacing   */
      static std::vector<hist2d>
      loglinear(unsigned, double, double, unsigned, double, double);
      
      /*   linear-logarithmic spacing   */
      static std::vector<hist2d>
      linearlog(unsigned, double, double, unsigned, double, double);
    };
  }   //  namespace distpoint
}   //   namespace duct





namespace duct {
  
  namespace __details {
    
    template<typename _Arg>
    inline std::ostream& distpoint_print(std::ostream& os, const _Arg& arg) {
      return os<<arg;
    }
    
    template<typename _Arg>
    requires std::is_signed<_Arg>::value
    inline std::ostream& distpoint_print(std::ostream& os, const _Arg& arg) {
      if(arg > 0) os<<" ";
      return os<<arg;
    }
    
    template<typename _Arg, typename ..._Args>
    inline std::ostream& distpoint_print(std::ostream& os, const _Arg& arg, const _Args& ...args) {
      distpoint_print(os, arg); os<<"  "; distpoint_print(os, args...);
      return os;
    }
    
    inline std::ostream&  distpoint_print(std::ostream& os,  const distpoint::hist1d& p) {
      return distpoint_print(os, p.xmin, 0.5*(p.xmin+p.xmax), p.xmax);
    }
    
    
    inline std::ostream&  distpoint_print(std::ostream& os,  const distpoint::hist2d& p) {
      return distpoint_print(os, p.xmin, 0.5*(p.xmin+p.xmax), p.xmax, p.ymin, 0.5*(p.ymin+p.ymax), p.ymax);
    }
  }  //   namespace __details
  
  
  
  /*   Default print for any dimension distributions    */
  template<typename _Point, typename _Tp>
  struct distprint
  {
    using _Stat = sample_statistic<_Tp>;
    
    /*    print one line of the distribution   */
    static std::ostream& print(std::ostream& os, const _Point& p, const _Stat& s) {
      return __details::distpoint_print(os, p)<<"  "<<s.mean()<<"  "<<s.error()<<'\n';
    }
    
    /*    print with normalization    */
    static std::ostream& print(std::ostream& os, const _Point& p, const _Stat& s, const _Stat& n) {
      auto norm = n.mean();
      return __details::distpoint_print(os, p)<<"  "<<s.mean()/norm<<"  "<<s.error()/norm<<"  "<<s.mean()<<"  "<<s.error()<<'\n';
    }
  };
  
  
  
  /*    Default print for arrays    */
  template<typename _Point, typename _Tp, std::size_t _Dim>
  struct distprint<_Point, std::array<_Tp, _Dim>>
  {
    /*   useful aliases    */
    using _Stat = sample_statistic<std::array<_Tp, _Dim>>;
    
    /*    print one line of the distribution   */
    static std::ostream& print(std::ostream& os, const _Point& p, const _Stat& x)
    {
      __details::distpoint_print(os, p);
      
      for(std::size_t k = 0; k < _Dim; k++)
        os<<"   "<<x.mean(k)<<" "<<x.error(k);
      
      return os<<'\n';
    }
    
    /*   printing the result with normalization */
    static std::ostream& print(std::ostream& os, const _Point& p, const _Stat& x, const _Stat& n)
    {
      __details::distpoint_print(os, p);
      
      /*   normalized distributions   */
      for(std::size_t k = 0; k < _Dim; k++) {
        auto norm = n.mean(k);
        os<<"   "<<x.mean(k)/norm<<" "<<x.error(k)/norm;
      }
      
      /*   non-normalized distributions    */
      for(std::size_t k = 0; k < _Dim; k++)
        os<<"   "<<x.mean(k)<<" "<<x.error(k);
      
      return os<<'\n';
    }
  };
  
  

  
  
  /*   Default print for zero dimension distributions    */
  template<typename _Tp>
  struct distprint<distpoint::nulld, _Tp>
  {
    using _Stat = sample_statistic<_Tp>;
    
    /*    print one line of the distribution   */
    static std::ostream& print(std::ostream& os, const _Stat& s) {
      return os<<s.mean()<<"  "<<s.error()<<'\n';
    }
  };
  
  
  
  /*    Default print for arrays with nulld poin type   */
  template<typename _Tp, std::size_t _Dim>
  struct distprint<distpoint::nulld, std::array<_Tp, _Dim>>
  {
    /*   useful aliases    */
    using _Stat = sample_statistic<std::array<_Tp, _Dim>>;
    
    /*    print one line of the distribution   */
    static std::ostream& print(std::ostream& os, const _Stat& x)
    {
      for(std::size_t k = 0; k < _Dim; k++)
        os<<"   "<<x.mean(k)<<" "<<x.error(k);
      
      return os<<'\n';
    }
  };
  
}   //   namespace duct







namespace duct {
  
  //   Forward declaration of the class distriburion
  template<typename _Tp, typename _Point = distpoint::nulld>
  class distribution;


  
  //
  //   Specialization : class _Point = void
  //
  template<typename _Tp>
  class distribution<_Tp, distpoint::nulld>
  {
  public:
    //       types
    using sample_type = _Tp;
    
    //   constructors
    distribution() { sample_statistic<_Tp>::zero(_M_tmp);}
    
    /*   computed assinment   */
    distribution& operator+=(const distribution& x) {
      _M_acc += x._M_acc;
      return *this;
    }

    /*   collecting weights   */
    void collect(const sample_type& weight) {
      sample_statistic<_Tp>::assadd(_M_tmp, weight);
    }
    
    /*     accumulate weights   */
    void accumulate() {
      _M_acc.accumulate(_M_tmp);
      reset_tmp();
    }
    
    void accumulate(const sample_type& weight) {
      sample_statistic<_Tp>::assadd(_M_tmp, weight);
      _M_acc.accumulate(_M_tmp);
      reset_tmp();
    }

    /*     reset   */
    void reset_tmp() { sample_statistic<_Tp>::zero(_M_tmp);}
    void reset() { _M_acc.reset(); reset_tmp();}
    
    /*  access to the results   */
    sample_statistic<_Tp>& get_accumulator() { return _M_acc;}
    const sample_statistic<_Tp>& get_accumulator() const { return _M_acc;}
    
    /*   no normalization for nulld histograms    */
    constexpr bool is_normalized() const { return false;}

  private:
    sample_type _M_tmp;
    sample_statistic<_Tp> _M_acc;
  };
  
  //   I/O functions for _Point = void
  template<typename _Tp>
  inline std::ostream& write(std::ostream& os, const distribution<_Tp, distpoint::nulld>& x) {
    return write(os, x.get_accumulator());
  }

  template<typename _Tp>
  inline std::istream& read(std::istream& is, distribution<_Tp, distpoint::nulld>& x)
  {
    read(is, x.get_accumulator());
    x.reset_tmp();
    return is;
  }
  
  template<typename _Tp>
  inline std::ostream& print(std::ostream& os, const distribution<_Tp, distpoint::nulld>& x) {
    return distprint<distpoint::nulld, _Tp>::print(os, x.get_accumulator());
  }


  //
  //    General template class 
  //
  template<typename _Tp, class _Point>
  class distribution
  {
  private:
    //    private types
    typedef distribution<_Tp, distpoint::nulld> _Dist;
    typedef std::pair<_Point, _Dist> _Value;
    typedef std::vector<_Value> _Base;
    
  public:
    //   public types
    typedef _Tp sample_type;
    typedef _Point point_type;

    typedef typename _Base::iterator iterator;
    typedef typename _Base::reference reference;
    typedef typename _Base::const_iterator const_iterator;
    typedef typename _Base::const_reference const_reference;
  
    typedef typename _Base::reverse_iterator reverse_iterator;
    typedef typename _Base::const_reverse_iterator const_reverse_iterator;
 
    /*   constructors, destructor   */
    distribution() : _M_isnorm(false) {}
    
    /*   constructors   */
    template<class _InputIterator>
    distribution(_InputIterator first, _InputIterator last, bool norm = false)
    : _M_isnorm(norm) {
      for(auto ib = first; ib != last; ++ib)
        _M_base.push_back(std::make_pair(*ib, _Dist{}));
    }
    
    distribution(unsigned int n, const _Point *base, bool norm = false)
    : distribution(base, base+n, norm) {}
        
    distribution(std::initializer_list<_Point> base, bool norm = false)
    : distribution(base.begin(), base.end(), norm) {}
    
    template<class _Container>
    explicit distribution(const _Container& base, bool norm = false)
    : distribution(base.begin(), base.end(), norm) {}
    
    
    /*   computed assignment   */
    distribution& operator+=(const distribution& x)
    {
      const_iterator ix = x._M_base.begin();
      for(iterator i = _M_base.begin(); i < _M_base.end(); i++, ix++)
        i->second += ix -> second;
      
      if(_M_isnorm && x._M_isnorm) _M_norm += x._M_norm;
      else if(x._M_isnorm) _M_norm = x._M_norm;
      
      _M_isnorm |= x._M_isnorm;
      
      return *this;
    }
   
    /*   collecting the weights    */
    template<class _Func>
    void collect(_Func func) {
      for(auto& is : _M_base)
        is.second.collect(func(is.first));
    }
    
    template<class _Func, typename ..._Arg>
    void collect(const _Tp& w, _Func&& func, _Arg&&... x)
    {
      for(auto& is : _M_base) {
        auto resf = func(is.first, std::forward<_Arg>(x)...);
        if(resf != 0.0) is.second.collect(sample_statistic<_Tp>::mul(w, resf));
      }
    }
    
    /*   accumulate the collected weights   */
    void accumulate() {
      for(auto& is : _M_base) is.second.accumulate();
    }
    
    template<class _Func>
    void accumulate(_Func&& func) {
      this->collect(std::forward<_Func>(func));
      this->accumulate();
    }
     
    template<class _Func, typename ..._Arg>
    void accumulate(const _Tp& w, _Func&& func, _Arg&&... x) {
      this->collect(w, std::forward<_Func>(func), std::forward<_Arg>(x)...);
      this->accumulate();
    }
    
    
    /*   collecting and accumulating weights for the normalization   */
    void collect_norm(const sample_type& weight) { _M_norm.collect(weight);}
    void accumulate_norm(const sample_type& weight) { _M_norm.accumulate(weight);}
    void accumulate_norm() { if(_M_isnorm) _M_norm.accumulate();}

    
    /*   reinitialize the distribution   */
    void reset() { for(auto& is : _M_base) is.second.reset(); _M_norm.reset();}
    void reset_tmp() { for(auto& is : _M_base) is.second.reset_tmp(); _M_norm.reset_tmp();}

    //  capacity
    unsigned int size() const { return _M_base.size();}
    void resize(unsigned int n) { _M_base.resize(n);}
    
    /*  iterators   */
    iterator begin() { return _M_base.begin();}
    iterator end() { return _M_base.end();}
    
    const_iterator begin() const { return _M_base.begin();}
    const_iterator end() const { return _M_base.end();}

    const_iterator cbegin() const { return _M_base.cbegin();}
    const_iterator cend() const { return _M_base.cend();}

    /*   reverse iterator   */
    reverse_iterator rbegin() { return _M_base.rbegin();}
    reverse_iterator rend() { return _M_base.rend();}
    
    const_reverse_iterator rbegin() const { return _M_base.rbegin();}
    const_reverse_iterator rend() const { return _M_base.rend();}
    
    const_reverse_iterator crbegin() const { return _M_base.crbegin();}
    const_reverse_iterator crend() const { return _M_base.crend();}
    
    /*   element access   */
    reference operator[](unsigned int i) { return _M_base[i];}
    const_reference operator[](unsigned int i) const { return _M_base[i];}
    
    /*   access to the normalization   */
    _Dist& norm() { return _M_norm;}
    const _Dist& norm() const { return _M_norm;}
    
    /*   check if it is normalized distribution   */
    bool is_normalized() const { return _M_isnorm;}
    
    /*   make it normalized distribution or reset normalization   */
    void normalize(const _Dist& nn = _Dist{}) { _M_norm = nn; _M_isnorm = true;}
   
    /*    I/O  operations   */
    void write(std::ostream& os) const {
      duct::write(os, _M_base, _M_isnorm, _M_norm);
    }

    void read(std::istream& is) {
      duct::read(is, _M_base, _M_isnorm, _M_norm);
      reset_tmp();
    }
   
  private:
    /*   data members   */
    _Base _M_base;
    
    /*    normalization  */
    bool _M_isnorm;
    _Dist _M_norm;
  };

    
  template<typename _Tp, class _Point>
  inline distribution<_Tp, _Point>
  operator+(const distribution<_Tp, _Point>& x, const distribution<_Tp, _Point>& y)
  {
    return distribution<_Tp, _Point>(x) += y;
  }
  
    
  //      I/O operations
  template<typename _Tp, class _Point>
  inline std::ostream& write(std::ostream& os, const distribution<_Tp, _Point>& d) {
    d.write(os);
    return os;
  }
  
  template<typename _Tp, class _Point>
  inline std::istream& read(std::istream& is, distribution<_Tp, _Point>& d) {
    d.read(is);
    return is;
  }

  template<class _Tp, typename _Point>
  std::ostream& print(std::ostream& os, const distribution<_Tp, _Point>& d)
  {
    const auto& norm = d.norm().get_accumulator();
    bool is_norm = d.is_normalized();
    
    for(auto& di : d)
      if(is_norm) distprint<_Point, _Tp>::print(os, di.first, di.second.get_accumulator(), norm);
      else distprint<_Point, _Tp>::print(os, di.first, di.second.get_accumulator());
    
    return os;
  }
  
}   //   namespace duct


#endif
