/*  Copyright (C) 2015 Zoltán Nagy
 *  Created by Zoltán Nagy on 3/21/15.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __mc_shower_history_h__
#define __mc_shower_history_h__ 1


/*   std includes   */


/*   project headers   */
#include <bits/duct-event.h>
#include <bits/duct-mapping.h>
#include <cxx/binary-tree.h>



namespace duct {
  
  namespace __details {
    
    /*   simple pair for returning the threshold weights   */
    struct thrs_weight
    {
      thrs_weight& operator+=(const thrs_weight& x) {
        weight_main += x.weight_main;
        weight_reg += x.weight_reg;
        weight_delta += x.weight_delta;
        return *this;
      }
      
      thrs_weight& operator*=(double x) {
        weight_main *= x;
        weight_reg *= x;
        weight_delta *= x;
        return *this;
      }
      
      double weight_main = 0.0, weight_reg = 0.0, weight_delta = 0.0;
    };
    
    
  }   //   namespace __details
  
  
  /*
   *    Simple class for providing the kT and virtuality veto parameters.
   */
  struct shower_vetos
  {
    /*  This is maximum allowed transverse momentum square in the initial
     *  and final state splitting. These usually set by the hard state.
     */
    double max_k2T = std::numeric_limits<double>::max();
    
    /*  This is maximum allowed virtuality in the initial
     *  state splitting. These usually set by the hard state.
     */
    double max_q2 = std::numeric_limits<double>::max();
  };
  
 
  /*   This class defines a set of weights those are carried through the shower evolution.   */
  template<class _Proc, bool _Is_there_ini = is_hadron_collision<_Proc>>
  struct weight_set;

  
  template<class _Proc>
  struct weight_set<_Proc, true>
  {
    /*    this is the weight type in the user function  */
    using user_weight_type = std::array<double, 4>;
    

    /*    user weight with LO PDF everywhere    */
    user_weight_type user_weight_lo_pdf(double fac = 1.0) const
    {
      std::array<double, 4> res;
      
      res[3] = unitary.main*fac;
      res[2] = res[3]*threshold.main;
      res[1] = res[2]*threshold.reg;
      res[0] = res[1]*threshold.delta;
      
      return res;
    }
    
    /*     default user weight is with NLO pdf    */
    user_weight_type user_weight_nlo_pdf(double fac = 1.0) const
    {
      std::array<double, 4> res;
      
      res[3] = unitary.main*fac*unitary.pdf_ratio;
      res[2] = res[3]*threshold.main;
      res[1] = res[2]*threshold.reg;
      res[0] = res[1]*threshold.delta;
      
      return res;
    }
    
    
    /*  some helper function for the shower engine   */
    void apply_threshold_correction(const __details::thrs_weight& thrs) {
      threshold.main *= std::exp(thrs.weight_main);
      threshold.reg *= std::exp(thrs.weight_reg);
      threshold.delta *= std::exp(thrs.weight_delta);
    }
    
    /*   weight of the unitary part of the shower   */
    struct {
      /*   this is the weight of the unitarized part   */
      double main;
      
      /*   ratio of NLO an LO MSbar PDFs at the hard state   */
      double pdf_ratio;
    } unitary;
    
    
    /*   threshold correction weights   */
    struct {
      /*   this is the main part of the threshold weight   */
      double main;
      
      /*   this is the contribution of the Preg(z) splitting kernels   */
      double reg;
      
      /*   this is the delta part of the threshold weight   */
      double delta;
    } threshold;
  };
  
  
  template<class _Proc>
  struct weight_set<_Proc, false>
  {
    /*    this is the weight type in the user function  */
    using user_weight_type = double;
    
    /*    calculates the weights for the user routine   */
    user_weight_type operator()(double fac = 1.0) const {
      return unitary.main*fac;
    }

    /*  some helper function for the shower engine (dummy)  */
    constexpr void apply_threshold_correction(const __details::thrs_weight&) {}
    
    /*   this is the weight of the unitarized part   */
    struct {
      /*   this is the weight of the unitarized part   */
      double main;
    } unitary;
  };
  


  
  /*  Class for recording every stage of the shower evolution.
   *  It consists of
   *   state         => the event
   *   scale         => scale of the event (equivalent to current time)
   *   weight        => weight of the current event
   *   vars          => spitting variable for the next stage
   *   qnull         => reference momentum for defining shower time
   *   current_time  => shower time when the event was created
   *   next_time     => shower time of the next splitting
   *   is_color_u3   => true if the stage is considered as U(3) state
   */
  template<class _Proc>
  struct shower_stage
  {
    /*    public types    */
    using event_type = event<_Proc>;
    
    
    /*  The event of the state. This contains every informations about the
     *  event, momenta, flavors, color connections, incoming objects,...
     */
    event_type state;
    
    /*   The scale of the shower stage. It is typically corresponds to the
     *   shower time, but at the hard state we don't have emitter parton and
     *   the definition of the shower time is ambigous...
     */
    double scale;
    
    /*  The posibble weight of the event. This weight can come from color
     *  factors or from spin correlation (when we add spin correlations).
     */
    weight_set<_Proc> weights;
    
    
    /*  Splitting varibles: This gives the splitting varibles of the next stage.
     *  It consists of delta, u, phi, emitter, spectator, color related variables.
     */
    splitting_vars vars;
   

    /*  This is the refernce momentum that help us to define shower time.
     *  We have a copy in very stage, since we want to resume shower from
     *  an intermediate state. In passive view this vector is boosted in
     *  every step of the shower.
     */
    lorentzvector<double> qnull;
    
    /*  The shower time when the event was created.
     */
    double current_time;
    
    /*  The shower time. This state is evolved without emission upto this time.
     */
    double next_time;

    /*   It is true if the state is consired as U(Nc) state.
     */
    bool is_color_u3;
    
    /*  It represents the color supression index of the state 
     *  that was accumulated during shower evolution.
     */
    unsigned int color_suppression_index;
    
    /*  Color supression index of the hard state which from the current state was obtained.
     *  It remains the same during the shower evolution.
     */
    unsigned int hard_color_suppression_index;
    
    
    /*   Since a history can have more than one hard state, we have to add 
     *   the veto parameters to this class. The won't change during the 
     *   evolution (I cannot imagine such a scenario... but who knows...).
     *   Maybe the upper limits on kT  or virtuality can change...
     *   Anyway, it is better to keep it here.
     */
    shower_vetos vetos;
  };
  
  
  /*   Shower history
   *   This is a big container class with a binary tree.
   *   Since it is a container we provide function only for 
   *   building up the history but we don't put the evolution 
   *   code here, like in the previous version.
   */
  
  template<class _Proc>
  class shower_history
  {
    /*   private types    */
    using _Tree = container::binary_tree<shower_stage<_Proc>, true>;
    using _TreeIter = typename _Tree::iterator;
    using _TreeConstIter = typename _Tree::const_iterator;
    
  public:
    /*   pulic types   */
    using stage_type = shower_stage<_Proc>;
    
    /*   cursors   */
    using cursor = typename _Tree::cursor;
    using const_cursor = typename _Tree::const_cursor;
    
    /*   ascending iterator, it steps to the next parent    */
    struct ascending : cursor
    {
      /*   some types to be compatible with the standard   */
      using value_type = stage_type;
      using pointer = stage_type*;
      using reference = stage_type&;
      
      using difference_type = std::ptrdiff_t;
      using iterator_category = std::forward_iterator_tag;
      
      /*   Constructors   */
      ascending() = default;
      
      /*   creating form const/non-const cursor type if it is possible   */
      explicit ascending(const cursor& c) : cursor{c} {}
      
      /*   Dereference operators   */
      reference operator*() const { return typename _Tree::iterator{*this}.operator*();}
      pointer operator->() const { return typename _Tree::iterator{*this}.operator->();}
      
      /*   Increment operators  */
      ascending& operator++() {
        cursor& c{*this};
        while(this->up().right() == c) c = this->up();
        c = this->up();
        return *this;
      }
      
      ascending operator++(int) {
        ascending tmp = *this;
        this->operator++();
        return tmp;
      }
      
      /*   check if the current stage is a hard state   */
      bool is_hard() const {
        ascending c{*this};
        return (++c).is_root();
      }
    };
    
    struct const_ascending : const_cursor
    {
      /*   some types to be compatible with the standard   */
      using value_type = const stage_type;
      using pointer = const stage_type*;
      using reference = const stage_type&;
      
      using difference_type = std::ptrdiff_t;
      using iterator_category = std::forward_iterator_tag;
      
      /*   constructors   */
      const_ascending() = default;
      explicit const_ascending(const cursor& c) : const_cursor{c} {}
      explicit const_ascending(const const_cursor& c) : const_cursor{c} {}
      
      /*   Dereference operators   */
      reference operator*() const { return typename _Tree::const_iterator{*this}.operator*();}
      pointer operator->() const { return typename _Tree::const_iterator{*this}.operator->();}
      
      /*   Increment operators  */
      const_ascending& operator++() {
        const_cursor& c{*this};
        while(this->up().right() == c) c = this->up();
        c = this->up();
        return *this;
      }
      
      const_ascending operator++(int) {
        const_ascending tmp = *this;
        this->operator++();
        return tmp;
      }
      
      /*   check if the current stage is a hard state   */
      bool is_hard() const {
        const_ascending c{*this};
        return (++c).is_root();
      }
    };
    
    
    /*  The iterator is the inorder traverser of the tree
     *  but we keep moving as long as the left node is not null.
     */
    struct iterator : _TreeIter
    {
      /*   Constructors   */
      iterator() = default;
      
      /*   creating form const/non-const cursor type if it is possible   */
      iterator(const cursor& c) : _TreeIter{c} {}
      
      /*   Increment operators  */
      iterator& operator++() {
        _TreeIter& c{*this};
        do ++c; while(c.left() != nullptr && !c.is_root());
        return *this;
      }
      
      iterator operator++(int) {
        iterator tmp = *this;
        this->operator++();
        return tmp;
      }
      
      /*   Decrement operators  */
      iterator& operator--() {
        _TreeIter& c{*this};
        do --c; while(c.left() != nullptr && !c.is_root());
        return *this;
      }
      
      iterator operator--(int) {
        iterator tmp = *this;
        this->operator--();
        return tmp;
      }
      
      /*   returns the corresponding hard event   */
      typename _TreeIter::reference
      hard() const {
        ascending c{*this}, h{*this};
        while(!c.is_root()) h = c++;
        return *h;
      }
      
      /*   check if the current pointee is a hard stage   */
      bool is_hard() const {
        return this->up().is_root();
      }
    };
    
    struct const_iterator : _TreeConstIter
    {
      /*   Constructors   */
      const_iterator() = default;
      
      /*   creating form const/non-const cursor type if it is possible   */
      const_iterator(const cursor& c) : _TreeConstIter{c} {}
      const_iterator(const const_cursor& c) : _TreeConstIter{c} {}
      
      /*   Increment operators  */
      const_iterator& operator++() {
        _TreeConstIter& c{*this};
        do ++c; while(c.left() != nullptr && !c.is_root());
        return *this;
      }
      
      const_iterator operator++(int) {
        const_iterator tmp = *this;
        this->operator++();
        return tmp;
      }
      
      /*   Decrement operators  */
      const_iterator& operator--() {
        _TreeConstIter& c{*this};
        do --c; while(c.left() != nullptr && !c.is_root());
        return *this;
      }
      
      const_iterator operator--(int) {
        const_iterator tmp = *this;
        this->operator--();
        return tmp;
      }
      
      /*   returns the corresponding hard event   */
      typename _TreeConstIter::reference
      hard() const {
        const_ascending c{*this}, h{*this};
        while(!c.is_root()) h = c++;
        return *h;
      }
      
      /*   check if the current pointee is a hard stage   */
      bool is_hard() const {
        return this->up().is_root();
      }
    };
    
    
    /*  This iterates over only the hard states.
     *  It is only forward iterator.
     */
    struct hard_iterator : cursor
    {
      /*   some types to be compatible with the standard   */
      using value_type = stage_type;
      using pointer = stage_type*;
      using reference = stage_type&;
      
      using difference_type = std::ptrdiff_t;
      using iterator_category = std::forward_iterator_tag;
      
      /*   Constructors   */
      hard_iterator() = default;
      
      /*   creating form const/non-const cursor type if it is possible   */
      explicit hard_iterator(const cursor& c) : cursor{c} {}
      
      /*   Dereference operators   */
      reference operator*() const { return typename _Tree::iterator{*this}.operator*();}
      pointer operator->() const { return typename _Tree::iterator{*this}.operator->();}
      
      /*   Increment operators  */
      hard_iterator& operator++() {
        cursor& c{*this};
        c = this->right();
        return *this;
      }
      
      hard_iterator operator++(int) {
        hard_iterator tmp{*this};
        this->operator++();
        return tmp;
      }
    };
    
    struct const_hard_iterator : const_cursor
    {
      /*   some types to be compatible with the standard   */
      using value_type = stage_type;
      using pointer = const stage_type*;
      using reference = const stage_type&;
      
      using difference_type = std::ptrdiff_t;
      using iterator_category = std::forward_iterator_tag;
      
      /*   Constructors   */
      const_hard_iterator() = default;
      
      /*   creating form const/non-const cursor type if it is possible   */
      explicit const_hard_iterator(const cursor& c) : const_cursor{c} {}
      explicit const_hard_iterator(const const_cursor& c) : const_cursor{c} {}
      
      /*   Dereference operators   */
      reference operator*() const { return typename _Tree::const_iterator{*this}.operator*();}
      pointer operator->() const { return typename _Tree::const_iterator{*this}.operator->();}
      
      /*   Increment operators  */
      const_hard_iterator& operator++() {
        const_cursor& c{*this};
        c = this->right();
        return *this;
      }
      
      const_hard_iterator operator++(int) {
        const_hard_iterator tmp{*this};
        this->operator++();
        return tmp;
      }
    };
    

    /*   default constructors    */
    shower_history()
    : _M_hist{} {}
    
    
    /*    standard range functions    */
    iterator begin() {
      _TreeIter c{_M_hist.begin()};
      while(c.left() != nullptr) ++c;
      return iterator{c};
    }
    
    const_iterator cbegin() const {
      _TreeConstIter c{_M_hist.cbegin()};
      while(c.left() != nullptr) ++c;
      return const_iterator{c};
    }
    
    const_iterator begin() const {
      return cbegin();
    }
    
    
    iterator end() { return iterator(_M_hist.end());}
    const_iterator end() const { return cend();}
    const_iterator cend() const { return const_iterator(_M_hist.cend());}
    
    
    /*    range of hard states    */
    iterator_range<hard_iterator> hard() {
      return {hard_iterator{_M_hist.head()}, hard_iterator{cursor{_M_hist.root(), nullptr}}};
    }
    
    iterator_range<const_hard_iterator> chard() const {
      return {const_hard_iterator{_M_hist.head()}, const_hard_iterator{const_cursor{_M_hist.root(), nullptr}}};
    }
    
    iterator_range<const_hard_iterator> hard() const {
      return chard();
    }
    
    /*    gives the history of a stage   */
    iterator_range<ascending> history_of(cursor c) {
      return {ascending{c}, ascending{_M_hist.root()}};
    }
    
    iterator_range<const_ascending> chistory_of(const_cursor c) const {
      return {const_ascending{c}, const_ascending{_M_hist.root()}};
    }
    
    iterator_range<const_ascending> history_of(const_cursor c) const {
      return chistory_of(c);
    }
    
    
    /*   start a new emission history    */
    void new_history() { _M_hist.clear();}
    
    /*   appending a bare operator on the given node
     *   we can append more than one operator on the same node
     */
    iterator append_bare(cursor root)
    {
      iterator n;
      
      if(root.left() == nullptr)
        n = iterator{_M_hist.append_left(root)};
      else {
        auto hard = root.left();
        while(hard.right() != nullptr) hard = hard.right();
        n = iterator{_M_hist.append_right(hard)};
      }
      
      /*   make sure that the vetos are relaxed (especially the kT and q2 cutoffs)    */
      n->vetos = shower_vetos{};
      
      return n;
    }
    
    iterator append_bare(cursor root, stage_type&& h)
    {
      iterator n;
      
      if(root.left() == nullptr)
        n = _M_hist.append_left(root, std::move(h));
      else {
        auto hard = root.left();
        while(hard.right() != nullptr) hard = hard.right();
        n = _M_hist.append_right(hard, std::move(h));
      }
      
      /*   it is not showered stage    */
      n->next_time = n->current_time;
      
      return n;
    }
    
    iterator append_bare(cursor root, const stage_type& h)
    {
      iterator n;
      
      if(root.left() == nullptr)
        n = _M_hist.append_left(root, h);
      else {
        auto hard = root.left();
        while(hard.right() != nullptr) hard = hard.right();
        n = _M_hist.append_right(hard, h);
      }
      
      /*   it is not showered stage    */
      n->next_time = n->current_time;
      
      return n;
    }
    
    /*   injecting the hard stages
     *   we can inject more than one hard states
     */
    iterator injecting_a_hard_stage() {
      return append_bare(_M_hist.root());
    }
    
    iterator injecting_a_hard_stage(stage_type&& h) {
      return append_bare(_M_hist.root(), std::move(h));
    }
    
    iterator injecting_a_hard_stage(const stage_type& h) {
      return append_bare(_M_hist.root(), h);
    }
    
    /*   append a shower stage  (it appends a node always to the left)
     *   we use these in the simple shower evolution
     */
    iterator append(cursor c) {
      return _M_hist.append_left(c);
    }
    
    iterator append(cursor c, stage_type&& h) {
      return _M_hist.append_left(c, std::move(h));
    }
    
    iterator append(cursor c, const stage_type& h) {
      return _M_hist.append_left(c, h);
    }
    
    /*   appending a new branch (we use in the tribble gluon vertex)
     *   we can append more than one operator on the same node
     */
    iterator append_branch(cursor root) {
      while(root.right() != nullptr) root = root.right();
      return _M_hist.append_right(root);
    }
   
    iterator append_branch(cursor root, stage_type&& h) {
      while(root.right() != nullptr) root = root.right();
      return _M_hist.append_right(root, std::move(h));
    }
    
    iterator append_branch(cursor root, const stage_type& h) {
      while(root.right() != nullptr) root = root.right();
      return _M_hist.append_right(root, h);
    }
    

  private:                                                         //   DATA MEMEBERS
    /*  I think it is better to delegate the underlying tree instead of
     *  inheriting it. This way we can control the acces to it.
     */
    _Tree _M_hist;
  };
  
 
 
}

#endif
