/*  Copyright (C) 2014 Zoltán Nagy
 *  Created by Zoltán Nagy on 2/5/14.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __mc_splitting_prob_h__
#define __mc_splitting_prob_h__ 1

/*   std includes   */

/*   project headers   */
#include <bits/mc-splitting-kernel.h>
#include <bits/mc-shower-history.h>
#include <bits/mc-engine-opts.h>



namespace duct {
  
  namespace __details {
    
    /*   simple class to discover the color connections in the event   */
    struct lcplus_partner
    {
      int connected;
      double weight;
    };
    
    
    /*  This routine counts the color connected parton of an emitter. It considers
     *  both the amplitude and conjugate amplitude. This function needs to do the
     *  shower evolution.
     */
    void counting_color_connections(stackvector<lcplus_partner, 4>&, int, const particle_state *, const model_ref&);
    
    template<class _Proc>
    inline void counting_color_connections(stackvector<lcplus_partner, 4>& spectators, int emitter, const event<_Proc>& p)
    {
      auto qcd = p.template data<qcd_out>();
      
      counting_color_connections(spectators, emitter, qcd.begin()-1, p.get_model());
    }

    
    /*    Functions to set up the list of hard_params<...> for initial and final state emitters.   */
    template<class _Ordering>
    struct parameters
    {
      /*    final state hard parameters   */
      template<class _Proc>
      static void init(std::vector<hard_params<fin>>&, const shower_stage<_Proc>&, const shower_engine_opts&);
      
      /*   initial state parameters    */
      static void init(std::array<hard_params<ini>,1>&, const shower_stage<dis>&, const shower_engine_opts&);
      static void init(std::array<hard_params<ini>,2>&, const shower_stage<hhc>&, const shower_engine_opts&);
    };
    
    
    
    template<class _Ordering>
    void parameters<_Ordering>::init(std::array<hard_params<ini>,1>& __pars, const shower_stage<dis>& stage, const shower_engine_opts& opts)
    {
      const event<dis>& p = stage.state;
      double s = p[-1].momentum*p[hadron{0}].momentum;
      hard_params<ini> &pars = __pars[0];
      
      /*   model and pdf tables   */
      pars.mdl = p.get_model();
      pars.aux = pars.mdl.aux(hadron{p[hadron{0}].key});
      
      /*   basic variables   */
      pars.emitter = 0;
      pars.fl = p[0].flavor;
      pars.eta = (p[0].momentum*p[-1].momentum)/s;
      pars.Q2 = 2.0*pars.eta*s;
      pars.mT = std::max({pars.mdl.alpha_qcd().q2min(), pars.aux->thrsT(pars.fl), sqr(opts.kT_cutoff)})/pars.Q2;
      
      /*   clear the spectator list   */
      pars.spectators.clear();
      
      /*  Spectator related parameters  */
      stackvector<lcplus_partner, 4> specs;
      counting_color_connections(specs, 0, p);
      
      for(const auto& s : specs)
      {
        double ska = p[s.connected].momentum*p[0].momentum;
        double skb = p[s.connected].momentum*p[-1].momentum;
        double m2k = p[s.connected].momentum.mag2();
        
        double zk = 1.0-2.0*skb/pars.Q2;
        double xi = skb/(ska+skb);
        double xT = std::sqrt(std::abs(1.0-m2k*pars.Q2/(4.0*ska*skb)));
        
        pars.spectators.push_back({s.connected, zk, xi, xT, s.weight});
      }
      
      /*  Variable to be able to map the shower time to v   */
      pars.vnull = ordering_traits<ini, _Ordering>::vnull(pars.Q2, stage.qnull, p[0].momentum);
      
      
      /*   veto parameters   */
      pars.mTmax = stage.vetos.max_k2T/pars.Q2;
      pars.mYmax = stage.vetos.max_q2/pars.Q2;
    }
    
    
    template<class _Ordering>
    void parameters<_Ordering>::init(std::array<hard_params<ini>,2>& __pars, const shower_stage<hhc>& stage, const shower_engine_opts& opts)
    {
      const event<hhc>& p = stage.state;
      hard_params<ini> *pars = __pars.data();

      double s = p[hadron{-1}].momentum*p[hadron{0}].momentum;
      double eta = (p[-1].momentum*p[hadron{0}].momentum)/s;
      double etb = (p[0].momentum*p[hadron{-1}].momentum)/s;
      double Q2 = 2.0*eta*etb*s;
      
      
      for(int ia : {-1,0})
      {
        int ib = (-1 == ia ? 0 : -1);
        
        /*   model and pdf tables   */
        pars->mdl = p.get_model();
        pars->aux = pars->mdl.aux(hadron{p[hadron{ia}].key});
        
        /*   basic variables   */
        pars->emitter = ia;
        pars->fl = p[ia].flavor;
        pars->Q2 = Q2;
        pars->eta = (-1 == ia ? eta : etb);
        pars->mT = std::max({pars->mdl.alpha_qcd().q2min(), pars->aux->thrsT(pars->fl), sqr(opts.kT_cutoff)})/Q2;

        /*   clear the spectator list   */
        pars->spectators.clear();
        
        /*  Spectator related parameters  */
        stackvector<lcplus_partner, 4> specs;
        counting_color_connections(specs, ia, p);
        
        for(const auto& s : specs)
        {
          double xi = 0.0, xT = 0.5, zk = 0.5;
          
          if(s.connected > 0) {
            double ska = (p[s.connected].momentum*p[ia].momentum);
            double skb = (p[s.connected].momentum*p[ib].momentum);
            double m2k = p[s.connected].momentum.mag2();
            
            zk = 1.0-2.0*skb/Q2;
            xi = skb/(ska+skb);
            xT = std::sqrt(std::abs(1.0-m2k*Q2/(4.0*ska*skb)));
          }
          
          pars->spectators.push_back({s.connected, zk, xi, xT, s.weight});
        }
        
        /*  Variable to be able to map the shower time to v   */
        pars->vnull = ordering_traits<ini, _Ordering>::vnull(Q2, stage.qnull, p[ia].momentum);
        
        /*   veto parameters   */
        pars->mTmax = stage.vetos.max_k2T/Q2;
        pars->mYmax = stage.vetos.max_q2/Q2;
        
        /*  next emitter   */
        ++pars;
      }
    }
    
    
    template<class _Ordering> template<class _Proc>
    void parameters<_Ordering>::init(std::vector<hard_params<fin>>& __pars, const shower_stage<_Proc>& stage, const shower_engine_opts& opts)
    {
      /*  Some simple information   */
      const event<_Proc>& p = stage.state;
      auto nparton = p.template number_of<qcd_out>();
      model_ref mdl = p.get_model();
      
      
      constexpr bool is_decay = std::is_same<_Proc, decay>::value;
      auto Q = p.template momentum_sum<incomings>();
      double Q2 = Q.mag2(), m2T = std::max({mdl.alpha_qcd().q2min(), sqr(opts.kT_cutoff)});
      
      /*   resize the vector of the hard parameters   */
      hard_params<fin> pars;
      
      __pars.resize(0);
      
      
      /*   loop over all the possible emitters    */
      for(int emitter = 1; emitter <= (int) nparton; ++emitter)
      {
        /*   is it decay?  &  model   */
        pars.is_decay = is_decay;
        pars.mdl = mdl;
        
        /*   basic parameters   */
        pars.emitter = emitter;
        pars.fl = p[emitter].flavor;
        
        pars.pQ = 2.0*(p[emitter].momentum*Q);
        pars.mT = m2T/pars.pQ;
        
        pars.al = Q2/pars.pQ;
        pars.bl = std::abs(p[emitter].momentum.mag2())/pars.pQ;
        pars.rl = std::sqrt(1.0-4.0*pars.al*pars.bl);
        pars.ymax = sqr((1.0-pars.bl)/(std::sqrt(pars.al)+std::sqrt(std::abs(pars.al+pars.bl-1.0)))) - pars.bl;
        
        /*   check if it is a special case   */
        bool is_special = (std::abs(1.0-pars.ymax/pars.rl) < 1.0e-5);
        
        auto pl = (2.0*pars.al/(1.0+pars.rl))*p[emitter].momentum;
        auto nl = Q - pl; pl -= ((1.0-pars.rl)/(1.0+pars.rl))*Q;
        
        /*   clear the spectator list   */
        pars.spectators.clear();
        
        /*  Spectator related parameters  */
        stackvector<lcplus_partner, 4> specs;
        counting_color_connections(specs, emitter, p);
        
        for(const auto& s : specs)
        {
          if(is_decay && 0 == s.connected)
            pars.spectators.push_back({s.connected, 0.5, 0.25, s.weight});
          else if(is_special && s.connected > 0)
            pars.spectators.push_back({s.connected, 1.0, 0.0, s.weight});
          else
          {
            double kQ = (Q*p[s.connected].momentum)*2.0*pars.rl/(1.0+pars.rl);
            double xi = (pl*p[s.connected].momentum)/kQ;
            double mu = std::abs(p[s.connected].momentum.mag2())*(pl*nl)/(kQ*kQ);
            
            pars.spectators.push_back({s.connected, xi, mu, s.weight});
          }
        }
        
        
        /*  Variable to be able to map the shower time to v   */
        pars.vnull = ordering_traits<fin, _Ordering>::vnull(pars.pQ, stage.qnull, p[emitter].momentum);
        
        /*   veto parameters   */
        pars.mTmax = stage.vetos.max_k2T/pars.pQ;
        pars.mYmax = stage.vetos.max_q2/pars.pQ;
        

        /*  next emitter   */
        __pars.push_back(pars);
      }
    }
 
    
    
    /*   Helper class for imlementing the splitting probabilities   */
    template<bool _Is_there_ini>
    struct splitting_prob_base;
    
    
    template<>
    struct splitting_prob_base<true>
    {
      /*   gauss integral to perform the tau integral in the threshold weight   */
      static const gauss_integral<10u> threshold_integralT;
      
      template<class _Ordering>
      static thrs_weight threshold(const shower_engine_opts&, const hard_params<ini> *, double);
    };
    
    
    template<>
    struct splitting_prob_base<false> {};
    
  }   /*   namespace __details    */
  
  
  
  /*   This class implements the splitting probabilities for various processes.
   */
  template<class _Proc, class _Ordering, bool _Is_there_ini = is_hadron_collision<_Proc>>
  class splitting_prob;
    
  
  template<class _Proc, class _Ordering>
  class splitting_prob<_Proc, _Ordering, true> : private __details::splitting_prob_base<true>
  {
    /*   number of the incoming hadrons    */
    static constexpr unsigned _N_ini = _Proc::number_of_incoming_hadrons;
   
    /*   private types   */
    using _Base = __details::splitting_prob_base<true>;
    
  public:
    /*   public types  */
    using process_type = _Proc;
    using ordering_type = _Ordering;
    
    /*    make the trial emission    */
    bool make_trial(const shower_engine_opts&, shower_stage<_Proc>&, random_engine_type&, double, double);
    
    /*   Calculates the threshold weights (aka K factor).
     *   This function should be invoked after make_trial(...). The function 
     *   make_trial initializes the list of hard parameters.
     */
    __details::thrs_weight threshold(const shower_engine_opts&, double, double) const;
   
    __details::thrs_weight threshold(const shower_engine_opts& opts, const shower_stage<_Proc>& stage)
    {
      /*   list of hard parameters for initial state    */
      __details::parameters<_Ordering>::init(_M_ini_params, stage, opts);
      
      /*   returns the threshold weight   */
      return threshold(opts, 9.9e+128, stage.current_time);
    }
    

  private:
    /*    list of hard_params for all the emitters   */
    std::array<hard_params<ini>, _N_ini> _M_ini_params;
    std::vector<hard_params<fin>> _M_fin_params;
    
    
    /*   splitting kernels   */
    splitting_kernel<ini, _Ordering> _M_ini_kernel;
    splitting_kernel<fin, _Ordering> _M_fin_kernel;
  };
  
  
  template<class _Proc, class _Ordering>
  class splitting_prob<_Proc, _Ordering, false> : private __details::splitting_prob_base<false>
  {
    /*   private types  */
    using _Base = __details::splitting_prob_base<false>;
  
  public:
    /*   public types  */
    using process_type = _Proc;
    using ordering_type = _Ordering;

    
    /*    make the trial emission    */
    bool make_trial(const shower_engine_opts&, shower_stage<_Proc>&, random_engine_type&, double, double);
    
    
    /*   calculates the threshold weights (aka K factor)    */
    constexpr __details::thrs_weight threshold(const shower_engine_opts&, const shower_stage<_Proc>&) {
      /*    The final state heavy flavor threshold effects are not yet implemented!
       *    This is a future project and I hope I can finish at some point....
       */
      return __details::thrs_weight{0.0, 0.0, 0.0};
    }
    
    
  private:
    /*    list of hard_params for all the emitters   */
    std::vector<hard_params<fin>> _M_params;

    
    /*   splitting kernels   */
    splitting_kernel<fin, _Ordering> _M_kernel;
  };
  
  

  template<class _Proc, class _Ordering>
  bool splitting_prob<_Proc, _Ordering, true>::
  make_trial(const shower_engine_opts& opts, shower_stage<_Proc>& stage, random_engine_type& rne, double tmU, double tmL)
  {
    bool is_possible = false;
    double tmT;
    
    
    /*   list of hard parameters for initial state    */
    stage.next_time = tmU;
    __details::parameters<_Ordering>::init(_M_ini_params, stage, opts);
    
    
    /*   try the initial state splittings    */
    for(const auto& pars : _M_ini_params)
    {
      if(!_M_ini_kernel.make_trial(tmT, std::addressof(stage.vars), rne, &pars, tmU, tmL)) continue;
      
      /*    splitting is possible    */
      tmU = tmT;
      is_possible = true;
    }
    
    
    /*   list of hard parameters for final state    */
    __details::parameters<_Ordering>::init(_M_fin_params, stage, opts);
    
    
    /*   try the final state splittings    */
    for(const auto& pars : _M_fin_params)
    {
      if(!_M_fin_kernel.make_trial(tmT, std::addressof(stage.vars), rne, &pars, tmU, tmL)) continue;
      
      /*    splitting is possible    */
      tmU = tmT;
      is_possible = true;
    }
    
    
    /*   next shower time   */
    if(is_possible)
      stage.next_time = tmU;

    return is_possible;
  }
  
  
  
  template<class _Proc, class _Ordering>
  bool splitting_prob<_Proc, _Ordering, false>::
  make_trial(const shower_engine_opts& opts, shower_stage<_Proc>& stage, random_engine_type& rne, double tmU, double tmL)
  {
    bool is_possible = false;
    double tmT;
    
    /*   list of hard parameters for final state    */
    stage.next_time = tmU;
    __details::parameters<_Ordering>::init(_M_params, stage, opts);
    
    
    /*   try the final state splittings    */
    for(const auto& pars : _M_params)
    {
      if(!_M_kernel.make_trial(tmT, std::addressof(stage.vars), rne, &pars, tmU, tmL)) continue;
      
      /*    splitting is possible    */
      tmU = tmT;
      is_possible = true;
    }
    
    
    /*   next shower time   */
    if(is_possible)
      stage.next_time = tmU;
    
    return is_possible;
  }
  
  

  template<class _Proc, class _Ordering>
  __details::thrs_weight splitting_prob<_Proc, _Ordering, true>::threshold(const shower_engine_opts& opts, double tmH, double tmL) const
  {
    const hard_params<ini> *pars = _M_ini_params.data();
    const qcd_coupling& as = pars->mdl.alpha_qcd();
    constexpr unsigned alpha = static_cast<unsigned>(_Ordering::alpha);
    
    /*  calculate the tau integral for the first incoming parton     */
    double vL, vH, tauL, tauH;
    __details::thrs_weight result{0.0, 0.0, 0.0};
    
    
    vH = std::min(pars->mTmax/std::pow(1.0-pars->eta, alpha), pars->vnull*std::exp(-tmL));
    vL = std::max(pars->mT, pars->vnull*std::exp(-tmH));
    
    if(vH > vL) {
      tauL = as.tauk(pars->Q2*vH);
      tauH = as.tauk(pars->Q2*vL);
      
      auto integ = _Base::threshold_integralT([=](double xi) -> __details::thrs_weight {
        return _Base::threshold<_Ordering>(opts, pars, (tauH-tauL)*0.5*(xi+1.0) + tauL);
      });
      
      integ *= 0.5*(tauH-tauL);
      result += integ;
    }
    
    
    /*   In hhc processes we have another incoming parton   */
    if(2u == _N_ini) {
      ++pars;
      
      vH = std::min(pars->mTmax/std::pow(1.0-pars->eta, alpha), pars->vnull*std::exp(-tmL));
      vL = std::max(pars->mT, pars->vnull*std::exp(-tmH));
      
      if(vH > vL) {
        tauL = as.tauk(pars->Q2*vH);
        tauH = as.tauk(pars->Q2*vL);
        
        auto integ = _Base::threshold_integralT([=](double xi) -> __details::thrs_weight {
          return _Base::threshold<_Ordering>(opts, pars, (tauH-tauL)*0.5*(xi+1.0) + tauL);
        });
        
        integ *= 0.5*(tauH-tauL);
        result += integ;
      }
    }
    
    return result;
  }

  

  
}

#endif
