/*  Copyright (C) 2013 Zoltán Nagy
 *  Created by Zoltán Nagy on 9/25/13.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __deductor_user_h__
#define __deductor_user_h__ 1


/*   std includes   */
#include <cstddef>
#include <cmath>
#include <initializer_list>
#include <algorithm>
#include <array>
#include <iostream>


/*   project headers   */
#include <bits/mc-basic-user.h>


/*    weight types    */
template<class _Tag>
using weight_t  = typename duct::weight_set<_Tag>::user_weight_type;


/*   This header is used only in the user modules. 
 *   I think it is harmless to put this here...
 */
using namespace std::literals::chrono_literals;




/*  This is a valarray like simple weight class to use in the user function as sample type
 *  With this we can have fancy ratio plot.
 */

template<std::size_t _Size>
class multiweight
{
public:
  /*   usual typedefs for an array   */
  using value_type = double;
  using pointer = double*;
  using const_pointer = const double*;
  using iterator = pointer;
  using const_iterator = const_pointer;
  
  /*    defaulted constructors and assignments    */
  multiweight() = default;
  multiweight(multiweight&&) = default;
  multiweight(const multiweight&) = default;
  ~multiweight() = default;
  
  multiweight& operator=(multiweight&&) = default;
  multiweight& operator=(const multiweight&) = default;
  
  
  /*    constructors    */
  explicit multiweight(double x) { std::fill(_M_val, _M_val+_Size, x);}
  multiweight(const double *p) { std::copy_n(p, _Size, _M_val);}
  
  
  /*   creating from initializer list    */
  multiweight(std::initializer_list<double> il) {
    auto min_size = std::min(_Size, il.size());
    std::copy_n(il.begin(), min_size, _M_val);
    if(min_size < _Size) std::fill(_M_val+il.size(), _M_val+_Size, *(il.end()-1));
  }
  
  /*    creating from standard array    */
  template<std::size_t _SizeA>
  explicit multiweight(const std::array<double, _SizeA>& p) {
    auto min_size = std::min(_Size, _SizeA);
    std::copy_n(p.begin(), min_size, _M_val);
    if(min_size < _Size) std::fill(_M_val+_SizeA, _M_val+_Size, p[_SizeA-1]);
  }
  
  /*    creating from C style array    */
  template<std::size_t _SizeA>
  explicit multiweight(const double (&p)[_SizeA]) {
    auto min_size = std::min(_Size, _SizeA);
    std::copy_n(p, min_size, _M_val);
    if(min_size < _Size) std::fill(_M_val+_SizeA, _M_val+_Size, p[_SizeA-1]);
  }
  
  template<std::size_t _SizeA>
  explicit multiweight(const multiweight<_SizeA>& p) {
    auto min_size = std::min(_Size, _SizeA);
    std::copy_n(p.begin(), min_size, _M_val);
    if(min_size < _Size) std::fill(_M_val+_SizeA, _M_val+_Size, p[_SizeA-1]);
  }
  
  
  /*    assignments   */
  multiweight& operator=(double x) {
    std::fill(_M_val, _M_val+_Size, value_type(x));
    return *this;
  }
  
  template<class _Up>
  multiweight& operator=(std::initializer_list<double> il)  {
    std::copy_n(il.begin(), std::min(_Size, il.size()), _M_val);
    return *this;
  }
  
  template<std::size_t _SizeA>
  multiweight& operator=(const multiweight<_SizeA>& p) {
    std::copy_n(p.begin(), std::min(_Size, _SizeA), _M_val);
    return *this;
  }
  
  template<std::size_t _SizeA>
  multiweight& operator=(const std::array<double, _SizeA>& p) {
    std::copy_n(p.begin(), std::min(_Size, _SizeA), _M_val);
    return *this;
  }
  
  template<std::size_t _SizeA>
  multiweight& operator=(const double (&p)[_SizeA]) {
    std::copy_n(p, std::min(_Size, _SizeA), _M_val);
    return *this;
  }
  
  
  /*   iterators   */
  constexpr iterator begin() { return _M_val;}
  constexpr iterator end() { return _M_val+_Size;}
  
  constexpr const_iterator begin() const { return _M_val;}
  constexpr const_iterator end() const { return _M_val+_Size;}
  
  constexpr const_iterator cbegin() const { return _M_val;}
  constexpr const_iterator cend() const { return _M_val+_Size;}
  
  
  /*   element acces   */
  constexpr value_type& operator[](std::size_t k) { return _M_val[k];}
  constexpr const value_type& operator[](std::size_t k) const { return _M_val[k];}
  
  /*   observers   */
  constexpr std::size_t size() const { return _Size;}
  
  
  /*   unary operators   */
  multiweight operator+() const {
    return *this;
  }
  
  multiweight operator-() const {
    multiweight res;
    std::transform(begin(), end(), res.begin(), [](double x) { return -x;});
    return res;
  }
  
  
  /*  computed assignment    */
  multiweight& operator*=(const value_type& x) {
    for(value_type& v : _M_val) v *= x;
    return *this;
  }
  
  multiweight& operator/=(const value_type& x) {
    for(value_type& v : _M_val) v /= x;
    return *this;
  }
  
  multiweight& operator+=(const value_type& x) {
    for(value_type& v : _M_val) v += x;
    return *this;
  }
  
  multiweight& operator-=(const value_type& x) {
    for(value_type& v : _M_val) v -= x;
    return *this;
  }
  
  
  multiweight& operator*=(const multiweight& v) {
    for(std::size_t k = 0; k < _Size; ++k) _M_val[k] *= v[k];
    return *this;
  }
  
  multiweight& operator/=(const multiweight& v) {
    for(std::size_t k = 0; k < _Size; ++k) _M_val[k] /= v[k];
    return *this;
  }
  
  multiweight& operator+=(const multiweight& v) {
    for(std::size_t k = 0; k < _Size; ++k) _M_val[k] += v[k];
    return *this;
  }
  
  multiweight& operator-=(const multiweight& v) {
    for(std::size_t k = 0; k < _Size; ++k) _M_val[k] -= v[k];
    return *this;
  }
  
  /*   binary I/O   */
  void read(std::istream& is) { duct::read(is, _M_val);}
  void write(std::ostream& os) const { duct::write(os, _M_val);}
  
private:
  /*   simple C array for the values   */
  value_type _M_val[_Size];
};


template<std::size_t _Size> inline
std::istream& read(std::istream& is, multiweight<_Size>& x) {
  x.read(is);
  return is;
}

template<std::size_t _Size> inline
std::ostream& write(std::ostream& os, const multiweight<_Size>& x) {
  x.write(os);
  return os;
}


template<std::size_t _Size> inline auto operator*(multiweight<_Size> a, const multiweight<_Size>& b) { return a *= b;}
template<std::size_t _Size> inline auto operator/(multiweight<_Size> a, const multiweight<_Size>& b) { return a /= b;}
template<std::size_t _Size> inline auto operator+(multiweight<_Size> a, const multiweight<_Size>& b) { return a += b;}
template<std::size_t _Size> inline auto operator-(multiweight<_Size> a, const multiweight<_Size>& b) { return a -= b;}


template<std::size_t _Size> inline auto operator*(multiweight<_Size> a, double b) { return a *= b;}
template<std::size_t _Size> inline auto operator/(multiweight<_Size> a, double b) { return a /= b;}
template<std::size_t _Size> inline auto operator+(multiweight<_Size> a, double b) { return a += b;}
template<std::size_t _Size> inline auto operator-(multiweight<_Size> a, double b) { return a -= b;}


template<std::size_t _Size> inline auto operator*(double b, multiweight<_Size> a) { return a *= b;}
template<std::size_t _Size> inline auto operator+(double b, multiweight<_Size> a) { return a += b;}


template<std::size_t _Size>
inline auto operator/(double b, const multiweight<_Size>& a)
{
  multiweight<_Size> r;
  for(std::size_t k = 0; k < _Size; ++k) r[k] = b/a[k];
  return r;
}

template<std::size_t _Size>
inline auto operator-(double b, const multiweight<_Size>& a)
{
  multiweight<_Size> r;
  for(std::size_t k = 0; k < _Size; ++k) r[k] = b-a[k];
  return r;
}


template<std::size_t _Size>
inline auto sqrt(const multiweight<_Size>& a)
{
  multiweight<_Size> r;
  for(std::size_t k = 0; k < _Size; ++k) r[k] = std::sqrt(a[k]);
  return r;
}


template<std::size_t _Size>
inline auto sqr(const multiweight<_Size>& a)
{
  multiweight<_Size> r;
  for(std::size_t k = 0; k < _Size; ++k) r[k] = a[k]*a[k];
  return r;
}


template<std::size_t _Size>
inline auto exp(const multiweight<_Size>& a)
{
  multiweight<_Size> r;
  for(std::size_t k = 0; k < _Size; ++k) r[k] = std::exp(a[k]);
  return r;
}

template<std::size_t _Size>
inline auto pow(const multiweight<_Size>& a, int p)
{
  multiweight<_Size> r;
  for(std::size_t k = 0; k < _Size; ++k) r[k] = std::pow(a[k], p);
  return r;
}

template<std::size_t _Size>
inline auto pow(const multiweight<_Size>& a, double p)
{
  multiweight<_Size> r;
  for(std::size_t k = 0; k < _Size; ++k) r[k] = std::pow(a[k], p);
  return r;
}

template<std::size_t _Size>
inline auto sin(const multiweight<_Size>& a)
{
  multiweight<_Size> r;
  for(std::size_t k = 0; k < _Size; ++k) r[k] = std::sin(a[k]);
  return r;
}

template<std::size_t _Size>
inline auto cos(const multiweight<_Size>& a)
{
  multiweight<_Size> r;
  for(std::size_t k = 0; k < _Size; ++k) r[k] = std::cos(a[k]);
  return r;
}


template<std::size_t _Size>
inline bool isfinite(const multiweight<_Size>& a)
{
  for(std::size_t k = 0; k < _Size; ++k)
    if(!std::isfinite(a[k])) return false;
 
  return true;
}



namespace duct {
  
  /*    Default print for multiweight    */
  template<typename _Point, std::size_t _Dim>
  struct distprint<_Point, multiweight<_Dim>>
  {
    /*   useful aliases    */
    using _Stat = sample_statistic<multiweight<_Dim>>;
    
    /*    print one line of the distribution   */
    static std::ostream& print(std::ostream& os, const _Point& p, const _Stat& x)
    {
      auto m = x.mean(), e = x.error();
      
      __details::distpoint_print(os, p);
      
      for(std::size_t k = 0; k < _Dim; k++)
        os<<"   "<<m[k]<<" "<<e[k];
      
      return os<<'\n';
    }
    
    /*   printing the result with normalization */
    static std::ostream& print(std::ostream& os, const _Point& p, const _Stat& x, const _Stat& n)
    {
      auto norm = n.mean();
      auto m = x.mean(), e = x.error();
      auto mn = m/norm, en = e/norm;
      
      __details::distpoint_print(os, p);
      
      /*   normalized distributions   */
      for(std::size_t k = 0; k < _Dim; k++)
        os<<"    "<<m[k]<<" "<<e[k];
      
      /*   non-normalized distributions    */
      for(std::size_t k = 0; k < _Dim; k++)
        os<<"    "<<mn[k]<<" "<<en[k];
      
      return os<<'\n';
    }
  };
  
 
  
  /*    Default print for multiweight with nulld point type   */
  template<std::size_t _Dim>
  struct distprint<distpoint::nulld, multiweight<_Dim>>
  {
    /*   useful aliases    */
    using _Stat = sample_statistic<multiweight<_Dim>>;
    
    /*    print one line of the distribution   */
    static std::ostream& print(std::ostream& os, const _Stat& x)
    {
      auto m = x.mean(), e = x.error();

      for(std::size_t k = 0; k < _Dim; k++)
        os<<"  "<<m[k]<<"  "<<e[k];
      
      return os<<'\n';
    }
  };


  /*    Default latex template for arrays    */
  template<typename _Point, std::size_t _Dim>
  requires std::is_same<_Point, distpoint::hist1d>::value || std::is_arithmetic<_Point>::value
  struct latex_template<_Point, multiweight<_Dim>>
  {
    /*    print a simple template of latex code to make the pretty plot  */
    static void plot(std::ostream& ltx, const std::string& title, const std::string& tblpath, bool is_normalized = false)
    {
      constexpr char xcol = std::is_same<_Point, distpoint::hist1d>::value ? '1' : '0';
      constexpr int c1 = std::is_same<_Point, distpoint::hist1d>::value ? 3 : 1;
      
      /*    writting the latex template file   */
      ltx<<"\n\\subsubsection*{\\it "<<title<<"}\n\n"
      <<"\\begin{tikzpicture}\n"
      <<"\\begin{axis}[title = {"<<title<<"},\n"
      <<"   xlabel={-- x label --}, ylabel={-- y label --},\n"
      <<"%  xmin=, xmax=,\n"
      <<"%  ymin=, ymax=\n]\n"
      <<"\\errorband[red,semithick]{fill=red!30!white, opacity=0.5}{"<<tblpath<<"}{"<<xcol<<"}{"<<c1<<"}{"<<c1+1<<"}\n\n";
      
      for(std::size_t i = 1; i < (is_normalized ? 2*_Dim : _Dim); ++i) {
        ltx<<"%\\errorband[blue,semithick]{fill=blue!30!white, opacity=0.5}{"<<tblpath<<"}{"<<xcol<<"}{"<<c1+2*i<<"}{"<<c1+1+2*i<<"}\n";
      }
      
      ltx
      <<"\\end{axis}\n"
      <<"\\end{tikzpicture}\n\n";
    }
  };
  

}   //  namespace duct

#endif
