/*  Copyright (C) 2013 Zoltán Nagy
 *  Created by Zoltán Nagy on 12/24/13.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */


/*   std headers   */
#include <cstring>
#include <string>
#include <iostream>
#include <fstream>
#include <list>
#include <set>

/*   C headers   */
#include <getopt.h>

/*   project headers   */
#include "cxx/filesystem.h"
#include "main-mod.h"


//----- print out the help -----
void main_mod_help()
{
  std::cout
  <<"Usage: deductor --module my-module --cflags=\"opt1 opt2 ...\"\n"
  <<"                                   --cxxflags=\"opt1 opt2 ...\"\n"
  <<"                                   --ldflags=\"opt1 opt2 ...\"\n"
  <<"                                   source1 source2 ...\n"
  <<"                                   -Lpath1 -llibrary1 -Lpath2 -llibrary2 ...\n"
  
  <<"\nOptioins of the --module mode:\n"
  <<"  my-module                        The name of the module. It will create the\n"
  <<"                                   loadable module file my-module.so.\n"
  <<"  --cflags=\"opt1 opt2 ...\"         C compiler options\n"
  <<"  --cxxflags=\"opt1 opt2 ...\"       C++ compiler options\n"
  <<"  --ldflags=\"opt1 opt2 ...\"        linker options\n"
  <<"  source1 source2 ...              C or C++ source files. Recognized extensions\n"
  <<"                                   are .c, .cc, .c++, .cxx, .cpp, .C\n"
  <<"  -Lpath                           library search path\n"
  <<"  -llibrary                        link the liblibrary.so (e.g.: -lfastjet)\n"
  <<"--------------------------------------------------------------------------------\n"
  <<std::endl;
}


namespace fs = duct::fs;


struct module_dsc
{

  /*   parsing the arguments  */
  std::set<fs::path> parse_args(int, char **);
  
  /*   write the assets file   */
  void write_assets(std::ostream&) const;
 
  /*   read the assets file    */
  void read_assets(std::istream&);
  
  /*   scan the scr forlder in the module   */
  void scan_src(const fs::path& moddir);

  /*   compiler and linker flags  */
  std::set<std::string> cflags, cxxflags, ldflags;

  /*  external libraries to link   */
  std::set<std::string> libs;
 
  /*  external library paths   */
  std::set<fs::path> lpath;
  
  /*  sources (C/C++ sources, headers and other files)   */
  std::set<fs::path> sources, others;
  
  /*  copy option when create new modulle or add filles to a module   */
  fs::copy_options cp_options = fs::copy_options::recursive | fs::copy_options::skip_existing;
  
private:
  /*   check if the string is a keyword   */
  static int is_keyword(const std::string& line)
  {
    static const char *keywords[] = {"CFLAG", "CXXFLAG", "LDFLAGS", "LIBS", "LIBPATHS", "SOURCES", "OTHERS"};
   
    for(int k = 0; k < 7; k++)
      if(std::strncmp(keywords[k], line.c_str(), std::strlen(keywords[k])) == 0)
        return k;
    
    return -1;
  }
};


std::set<fs::path> module_dsc::parse_args(int argc, char **argv)
{
  /*   long options  */
  ::option longopt[] = {
    {"cflags",   required_argument, nullptr, 0},
    {"cxxflags", required_argument, nullptr, 0},
    {"ldflags",  required_argument, nullptr, 0},
    {"update",   no_argument, nullptr, 0},
    {nullptr, 0, nullptr, 0}
  };
  
  /*   argument parsing   */
  while(true) {
    int optidx = 0;
    int c = ::getopt_long(argc, argv, "L:l:", longopt, &optidx);
    if(c == -1) break;
    
    switch(c) {
      case 'L': lpath.emplace(::optarg); break;
      case 'l': libs.emplace(::optarg); break;
      case 0:
        switch(optidx) {
          case 0: cflags.emplace(::optarg); break;
          case 1: cxxflags.emplace(::optarg); break;
          case 2: ldflags.emplace(::optarg);  break;
          case 3: cp_options = fs::copy_options::recursive | fs::copy_options::update_existing; break;
        }
        break;
      case '?': break;
      default: throw; break;
    }
  }
  
  /*   making the list of sources and other files   */
  std::set<fs::path> files;
  
  for(int k = ::optind; k < argc; k++)
    files.emplace(argv[k]);

  return files;
}


void module_dsc::write_assets(std::ostream& os) const
{
  os<<"CFLAGS:\n";
  for(const auto& f : cflags)
    os<<"  "<<f<<"\n";
  
  os<<"CXXFLAGS:\n";
  for(const auto& f : cxxflags)
    os<<"  "<<f<<"\n";
  
  os<<"LDFLAGS:\n";
  for(const auto& f : ldflags)
    os<<"  "<<f<<"\n";
  
  os<<"LIBS:\n";
  for(const auto& f : libs)
    os<<"  "<<f<<"\n";
  
  os<<"LIBPATHS:\n";
  for(const auto& f : lpath)
    os<<"  "<<f.string()<<"\n";
  
  os<<"SOURCES:\n";
  for(const auto& f : sources)
    os<<"  "<<f.string()<<"\n";
  
  os<<"OTHERS:\n";
  for(const auto& f : others)
    os<<"  "<<f.string()<<"\n";
}


void module_dsc::read_assets(std::istream& is)
{
  std::string line;
  int section = -1, is_key;
  
  while(std::getline(is>>std::ws, line)) {
    is_key = is_keyword(line);
    
    if(is_key == -1)
      switch(section) {
        case 0: cflags.insert(line); break;
        case 1: cxxflags.insert(line); break;
        case 2: ldflags.insert(line); break;
        case 3: libs.insert(line); break;
        case 4: lpath.insert(line); break;
        case 5: sources.insert(line); break;
        case 6: others.insert(line); break;
      }
    else section = is_key;
  }
}


void module_dsc::scan_src(const fs::path& moddir)
{
  /*   sorce directory   */
  fs::path srcdir = moddir/"src";
  auto n = srcdir.string().size();
  
  /*  scanning the source directory recursively  */
  for(const auto& f : fs::recursive_directory_iterator(srcdir))
  {
    if(is_regular_file(f) || is_symlink(f)) {
      /*   erasing the srcdir part rol the filename   */
      auto sf = f.path().string();
      sf.erase(0, n+1);
      
      /*  selecting source files   */
      auto ext = f.path().extension();
      
      if(ext == ".c" || ext == ".cc" || ext == ".cpp" || ext == ".cxx" || ext == ".c++" || ext == ".C")
        sources.emplace(sf);
      else others.emplace(sf);
    }
  }
}


static void mod_new(const fs::path& moddir, int argc, char **argv)
{
  /*    create the module directories           */
  fs::path srcdir = moddir/"src";
  
  fs::create_directories(srcdir);
  
  /*   parsing the command line   */
  module_dsc mod;
  
  auto files = mod.parse_args(argc, argv);
  
  /*   adding default flags and libs  */
  mod.cflags.emplace(__CFLAGS__);
  mod.cxxflags.emplace(__CXXFLAGS__);
  mod.libs.emplace("deductor-proc");
  mod.libs.emplace("deductor");
  mod.libs.emplace("ductcxx");
  mod.lpath.emplace(__LIBDIR__);
  
  /*  copying files to the src directory   */
  for(const auto& f : files) {
    /*   create the subfolders if it is neaded  */
    fs::create_directories((srcdir/f).parent_path());
    
    /*   copy files into the module src folder   */
    fs::copy(f, srcdir/f);
  }
  
  mod.scan_src(moddir);
  
  /*   write file assets    */
  std::fstream assets((srcdir/".assets").c_str(), std::ios::out);
  
  mod.write_assets(assets);
  assets.flush();
  assets.close();
  
  /*   write the summary to the screen   */
  mod.write_assets(std::cout);
}



static void mod_add(const fs::path& moddir, int argc, char **argv)
{
  /*   read back the assets file   */
  module_dsc mod;
  fs::path srcdir = moddir/"src";
  std::fstream assets((srcdir/".assets").c_str(), std::ios::in);
  
  mod.read_assets(assets);
  assets.flush();
  assets.close();
  
  /*   parsing the command line   */
  auto files = mod.parse_args(argc, argv);

  /*  copying files to the src directory   */
  for(const auto& f : files) {
    /*   create the subfolders if it is neaded  */
    fs::create_directories((srcdir/f).parent_path());
    
    /*   copy files into the module src folder   */
    fs::copy(f, srcdir/f, mod.cp_options);
  }
  
  mod.scan_src(moddir);
  
  /*   write file assets    */
  assets.open((srcdir/".assets").c_str(), std::ios::out);
  
  mod.write_assets(assets);
  assets.flush();
  assets.close();
  
  /*   write the summary to the screen   */
  mod.write_assets(std::cout);
}


static void mod_bld(const fs::path& moddir)
{
  /*   read back the assets file   */
  module_dsc mod;
  fs::path srcdir = moddir/"src";
  std::fstream file((srcdir/".assets").c_str(), std::ios::in);
  
  mod.read_assets(file);
  file.flush();
  file.close();
  
  /*  creating build dir if it is not there   */
  fs::path blddir = moddir/".build";
  fs::create_directories(blddir);
  
  /*   writing out the makefile    */
  file.open((blddir/"CMakeLists.txt").c_str(), std::ios::out);
 
  /*   making flag strings   */
  std::string cflags, cxxflags, ldflags, libs, lpath;
  
  for(const auto& flag : mod.cflags) {
    cflags += " ";
    cflags += flag;
  }

  for(const auto& flag : mod.cxxflags) {
    cxxflags += " ";
    cxxflags += flag;
  }

  for(const auto& flag : mod.ldflags) {
    ldflags += " ";
    ldflags += flag;
  }

  for(const auto& lib : mod.libs) {
    libs += " ";
    libs += lib;
  }
 
  for(const auto& path : mod.lpath) {
    lpath += " \"";
    lpath += path.string();
    lpath += "\"";
  }
  
  /*   writting the file   */
  file<<"cmake_minimum_required(VERSION 2.8)\n"
  <<"set(CMAKE_C_COMPILER "<<"\""<<__CC__<<"\")\n"
  <<"set(CMAKE_CXX_COMPILER "<<"\""<<__CXX__<<"\")\n"
  <<"project(\"module\")\n\n"
  
  <<"set(CMAKE_INSTALL_PREFIX ..)\n"
  <<"set(CMAKE_C_FLAGS \"${CMAKE_C_FLAGS} "<<cflags<<"\")\n"
  <<"set(CMAKE_CXX_FLAGS \"${CMAKE_CXX_FLAGS} "<<cxxflags<<"\")\n"
  <<"set(CMAKE_MODULE_LINKER_FLAGS \"${CMAKE_MODULE_LINKER_FLAGS}"<<ldflags<<"\")\n\n"
  
  <<"include_directories(\""<<__INCLUDEDIR__<<"\")\n"
  <<"link_directories("<<lpath<<")\n\n";
  
  file<<"add_library(module MODULE\n";
  for(const auto& src : mod.sources)
    file<<"../src/"<<src.string()<<"\n";
  file<<")\n\n";
  
  file<<"target_link_libraries(module "<<libs<<")\n";
  file<<"set_target_properties(module PROPERTIES PREFIX \"\")\n";
  file<<"install(TARGETS module DESTINATION .)\n";
  
  file.flush();
  file.close();

  /*   change into the bundle   */
  fs::current_path(moddir/".build/");
  
  /*   rum cmake and make   */
  std::string cmd = __CMAKE__;
  cmd += " -G \"Unix Makefiles\" .; ";
  cmd += __MAKE__;
  cmd += " install";
  
  std::system(cmd.c_str());
}


static void mod_cln(const fs::path& moddir, const fs::path& newmod)
{
  /*    => source directory                 */
  fs::create_directories((newmod/"src").c_str());
  
  /*  copying the src directory   */
  fs::copy(moddir/"src", newmod/"src", fs::copy_options::recursive);
  
}



void main_mod(char mode, int argc, char **argv)
{
  /*   creating the module build directory  */
  fs::path moddir = std::string{argv[::optind-1]}+".bundle/";
  
  /*   checking if the module exists   */
  bool is_there = exists(moddir);
  
  /*   create a new or add files and/or options   */
  if('n' == mode || 'a' == mode || 'm' == mode) {
    if(is_there) mod_add(moddir, argc, argv);
    else mod_new(moddir, argc, argv);
    
    /*   build if it was --module   */
    if('m' == mode) mod_bld(moddir);
      
    return;
  }
  
  /*   build the module   */
  if('b' == mode && is_there) {
    mod_bld(moddir);
    return;
  }
  
  /*   clone the module   */
  if('c' == mode) {
    /*   new module name   */
    fs::path newmod = std::string{argv[::optind]}+".bundle/";
    
    /*    clone the module it drops an error if it exists */
    mod_cln(moddir, newmod);
    return;
  }
  
}


