/*  Copyright (C) 2013 Zoltán Nagy
 *  Created by Zoltán Nagy on 9/8/13.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

/*   std headers   */
#include <string>
#include <utility>
#include <fstream>
#include <list>
#include <map>
#include <thread>
#include <chrono>

/*   C headers   */
#include <dlfcn.h>
#include <getopt.h>

/*   project headers   */
#include "main-mod.h"
#include "cxx/binary-io.h"
#include "cxx/filesystem.h"
#include "bits/duct-module.h"
#include "bits/conf-deductor.h"


/*     This is the main function. Now I can do this....   */
using namespace duct;
using namespace std;
using namespace std::chrono_literals;


std::string getfuncname(const fs::path& fn)
{
  std::fstream is(fn.c_str(), std::ios::in);

  /*   type of the saved file   */
  std::string dummy, fname;
  
  /*   reading number of events and typeinfo   */
  read(is, dummy, fname);
  
  is.close();
  
  return fname;
}


void latex_summary(std::ostream&, const std::list<std::string>&, const fs::path&);



int main_res(int argc, char **argv)
{
  /*   module path   */
  std::string modname{argv[::optind-1]};
  
  /*   parsing the command line   */
  ::option longopt[] = {
    {"daemon",   no_argument, nullptr, 0},
    {nullptr, 0, nullptr, 0}
  };
  
  
  /*   argument parsing   */
  bool is_daemon = false;
  
  while(true) {
    int optidx = 0;
    int c = ::getopt_long(argc, argv, "", longopt, &optidx);
    if(c == -1) break;
    
    switch(c) {
      case 0:
        switch(optidx) {
          case 0: is_daemon = true; break;
        }
        break;
      case '?': break;
      default: throw; break;
    }
  }
  
  
  /*   outputdir and resultdir path   */
  fs::path outdir = std::string{modname}+".bundle/output";
  fs::path resdir = std::string{modname}+".bundle/result";

  
  /*   opening the module    */
  cout<<"\033[1mOpening module \033[1;32m\""<<modname<<"\"\033[0m...\n";
  
  fs::path modfile = std::string{modname}+".bundle/module"+__SOEXT__;
  void *handle = ::dlopen(modfile.c_str(), RTLD_LAZY);
  
  if(!handle) {
    std::cerr<<"Unable to open the module \""<<modname<<"\"!\n";
    throw dlerror();
  }

  
  /*   run the main part   */
  do
  {
    /*  check if we have already output dir   */
    while(!fs::is_directory(outdir)) {
      if(is_daemon) std::this_thread::sleep_for(1.0min);
      else return 0;
    }
    
    auto n = outdir.string().size();
    
    /*   type of the collect and print functions    */
    using cpfunc_t = unsigned long long (*)(const fs::path&, const std::list<fs::path>&);
    
    /*   container to sort out the files   */
    std::map<std::string, std::pair<std::list<fs::path>, cpfunc_t>> files;
    
    /*   reading the output directory   */
    for(const fs::path& f : fs::recursive_directory_iterator(outdir))
      if(fs::is_regular_file(f) && f.extension() == ".dat") {
        /*   erasing the outdir part of the filename   */
        auto sf = f.parent_path().string();
        sf.erase(0, n+1);
        
        files[sf].first.push_back(f);
        files[sf].second = nullptr;
      }
    
    
    /*   finding the collect and print functions     */
    for(auto& u : files)
    {
      /*   checking the function name in the data files   */
      std::string func = getfuncname(u.second.first.front());
      
      for(const auto& f : u.second.first)
        if(func != getfuncname(f)) {
          std::cerr
          <<"  \033[1;31mError:\033[0m There is some problem with the data files\n"
          <<"         for analyser \"\033[1;31m"<<u.first<<"\033[0m\".\n"
          <<"         This analyser will be ignored. Please check your module!\n";
          
          goto __next_module;
        }
      
      /*   find the collect and print function in the module   */
      u.second.second = reinterpret_cast<cpfunc_t>(::dlsym(handle, func.c_str()));
      
      if(u.second.second == nullptr) {
        std::cerr
        <<"  \033[1;31mError:\033[0m Unable to locate the \"collect_and_print(...)\" function\n"
        <<"         for analyser \"\033[1;31m"<<u.first<<"\033[0m\" in the module.\n"
        <<"         This analyser will be ignored. Please check your module!\n";
        
        goto __next_module;
      }
      
      /*   generate the result in human readable format    */
      std::cout
      <<"Analyser \"\033[1;31m"<<u.first<<"\033[0m\" has collected \033[1m"
      <<(*u.second.second)(resdir/u.first, u.second.first)<<"\033[0m MC events.\n";
      
      
    __next_module:
      /*   empty statement   */;
    }
    
    /*    creating the latex summary     */
    fs::path texfile = resdir/"summary.tex";
    if(fs::exists(texfile)) texfile = resdir/"summary-default.tex";
    
    /*   writing the new summary file   */
    std::list<std::string> analysers;
    
    for(auto& u : files)
      analysers.push_back(u.first);
    
    std::fstream file(texfile.c_str(), std::ios::out);
    latex_summary(file, analysers, resdir);
    file.close();
  
    if(is_daemon) std::this_thread::sleep_for(2.0min);
  } while(is_daemon);
  
  
  
  /*   closing the module   */
  ::dlclose(handle);
  
  return 0;
}


void latex_summary(std::ostream& file, const std::list<std::string>& files, const fs::path& resdir)
{
  /*    printing the head of the latex file    */
  file<<R"tex(\documentclass[twoside,a4paper,11pt]{article}
\usepackage{jheppub, pgfplots,pgfplotstable,verbatim}
\usepgfplotslibrary{fillbetween}
\pgfplotsset{compat=newest}
  
  
%  simple macro to plot errorbands
\newcommand{\errorband}[6][]{
  \pgfplotstableread{#3}\datatable
  \addplot[name path=pluserror,draw=none,no markers,forget plot]  table [x={#4},y expr=\thisrow{#5}+\thisrow{#6}]{\datatable};
  \addplot[name path=minuserror,draw=none,no markers,forget plot] table [x={#4},y expr=\thisrow{#5}-\thisrow{#6}]{\datatable};
  \addplot[forget plot,#2] fill between[on layer={},of=pluserror and minuserror];
  \addplot [#1,no markers] table [x={#4},y={#5}]{\datatable};
}


%   useful macro to make ratio plots
\newcommand{\ratioplot}[8][]{
  \pgfplotstableread{#3}\datatable
    
  \addplot[name path=pluserror,draw=none,no markers,forget plot]  table
  [ x={#4},
    y expr=\thisrow{#5}/\thisrow{#7} + abs(\thisrow{#5}/\thisrow{#7})*(\thisrow{#6}/\thisrow{#5}+\thisrow{#8}/\thisrow{#7})
  ]{\datatable};
    
  \addplot[name path=minuserror,draw=none,no markers,forget plot]  table
  [ x={#4},
    y expr=\thisrow{#5}/\thisrow{#7}-abs(\thisrow{#5}/\thisrow{#7})*(\thisrow{#6}/\thisrow{#5}+\thisrow{#8}/\thisrow{#7})
  ]{\datatable};
    
  \addplot[forget plot,#2] fill between[on layer={},of=pluserror and minuserror];
  \addplot [#1,no markers] table [x={#4},y expr=\thisrow{#5}/\thisrow{#7}]{\datatable};
}

  
%  set the default style
\pgfplotsset{every axis/.style={
    width=12cm,
    height=10cm,
    grid=both,
    scaled ticks=false,
    yticklabel style={/pgf/number format/.cd, fixed,precision=5}
  }
}
  
%%%%%%%%%%%%%%%%%&%%%%%%%%%% TITLE PAGE DETAILS %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  
\title{Your wonderful calculations}
  
\author[a]{Your Name}
\author[b]{and Your Colleague}
  
\affiliation[a]{
  Your Institute\\
  Your Address
}
  
\affiliation[b]{
  His Institute\\
  His Address
}
  
\emailAdd{Your@Email.address}
\emailAdd{His@Email.address}
  

\abstract{
    These are wonderful calculations because ....
}
  
\keywords{perturbative QCD, parton shower}
\preprint{\today}
\dedicated{This document was created by {\sc Deductor} version )tex"<<conf::version<<".}\n\n"
<<R"tex(%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  
\begin{document}
\maketitle

\section{Introduction}
  
)tex";
  
  /*    printing the results    */
  std::string curr[3], mod, proc;
  
  for(auto& u : files)
  {
    int i = 0;
    for(auto&& p : fs::path{u}) {
      curr[i++] = p.string();
      if(i > 2) break;
    }
    
    /*   Every module in a new section    */
    if(mod != curr[0]) {
      file<<"\n\\section{Module: \\tt{"<<curr[0]<<"}}\n\n";
      mod = curr[0];
    }
    
    /*   Every process in a new subsection   */
    if(proc != curr[1]) {
      file<<"\\subsection{Process: \\tt{"<<curr[1]<<"}}\n\n";
      proc = curr[1];
    }
    
    /*   Every analyser in a new subsubsection   */
    file<<"\\subsubsection{Analyser: \\tt{"<<curr[2]<<"}}     "
    <<"%%% PATH: "<<u<<"\n"
    <<"\\def\\mypath{\""<<u<<"\"}\n"
    <<"\\input{\\mypath/analyser-info.tex}\n\n"
    <<"The number of collected MC events is \\mcevent.\n\n";
    
    /*   reading the output directory   */
    char c;
    std::fstream ltx;

    for(const fs::path& f : fs::recursive_directory_iterator(resdir/u))
      if(fs::is_regular_file(f) && f.filename() == "latex-template.tex") {
        ltx.open(f.c_str(), std::ios::in);
        while(ltx.get(c)) file.put(c);
        ltx.flush();
        ltx.close();
      }
    
  }
  
  /*   references   */
  file<<R"tex(\begin{thebibliography}{99}
  \bibitem{DeductorCode}
  {\sc Deductor} is available at
  \href{http://www.desy.de/~znagy/deductor/}
  {http://www.desy.de/$\sim$znagy/deductor/}
  and
  \href{http://pages.uoregon.edu/soper/deductor/}
  {http://pages.uoregon.edu/soper/deductor/}.
          
  \bibitem{NSI}
  Z.~Nagy and D.~E.~Soper,
  {\em Parton showers with quantum interference},
  JHEP {\bf 0709} (2007) 114
  [arXiv:0706.0017]]
  \href{http://inspirehep.net/record/752058/}{[INSPIRE]}.
  %%CITATION = JHEPA,0709,114;%%
    
  \bibitem{NSII}
  Z.~Nagy and D.~E.~Soper,
  {\em Parton showers with quantum interference: leading color, spin averaged},
  JHEP {\bf 0803} (2008) 030
  [arXiv:0801.1917]
  \href{http://inspirehep.net/record/777310/}{[INSPIRE]}.
  %%CITATION = JHEPA,0803,030;%%
      
  \bibitem{NSspin}
  Z.~Nagy and D.~E.~Soper,
  {\em Parton showers with quantum interference: leading color, with spin},
  [arXiv:0805.0216]
  \href{http://inspirehep.net/record/784889/}{[INSPIRE]}.
  %%CITATION = ARXIV:0805.0216;%%

  \bibitem{NScolor}
  Z.~Nagy and D.~E.~Soper,
  {\em Parton shower evolution with subleading color},
  JHEP {\bf 1206} (2012) 044
  [arXiv:1202.4496]
  \href{http://inspirehep.net/record/1089845?ln=en}{[INSPIRE]}.
  %%CITATION = ARXIV:1202.4496;%%
      
  \bibitem{deductor}
  Z.~Nagy and D.~E.~Soper,
  {\em A parton shower based on factorization of the quantum density matrix},
  JHEP {\bf 1406} (2014) 097
  [arXiv:1401.6364]
  \href{http://inspirehep.net/record/1278744?ln=en}{[INSPIRE]}.
  %%CITATION = ARXIV:1401.6364;%%
        
  \bibitem{ShowerTime}
  Z.~Nagy and D.~E.~Soper,
  {\em Ordering variable for parton showers},
  JHEP {\bf 1406} (2014) 178
  [arXiv:1401.6366]
  \href{http://inspirehep.net/record/1278745?ln=en}{[INSPIRE]}.
  %%CITATION = ARXIV:1401.6366;%%
                        
  \bibitem{MassivePdfs}
  Z.~Nagy and D.~E.~Soper,
  {\em Parton distribution functions in the context of parton showers},
  JHEP {\bf 1406} (2014) 179
  [arXiv:1401.6368]
  \href{http://inspirehep.net/record/1278746?ln=en}{[INSPIRE]}.
  %%CITATION = ARXIV:1401.6368;%%
\end{thebibliography}
)tex";

  
  /*   end document   */
  file<<"\\end{document}\n";
}




//----- print out the help -----
void main_res_help()
{
  std::cout
  <<"Usage: deductor --result modname [--daemon]\n\n"
  
  <<"Options of --add mode:\n"
  <<"  my-module            The module file contains the user defined functions.\n"
  <<"  -r directory         Save the result to this directory (default: ./result)\n"
  <<"--------------------------------------------------------------------------------\n"
  <<std::endl;
}


